""":class:`DataContainer` class implementation.
"""
import numpy as np

class dict_to_object:
    def __init__(self, finstance):
        self.finstance = finstance

    def __get__(self, instance, cls):
        return return_obj_method(self.finstance.__get__(instance, cls), instance, cls)

class return_obj_method:
    sig_attrs = {'__annotations__', '__doc__', '__module__',
                 '__name__', '__qualname__'}

    def __init__(self, func, instance, cls):
        self.instance, self.cls = instance, cls
        self.__wrapped__ = func
        for attr in self.sig_attrs:
            self.__dict__[attr] = getattr(func, attr)

    def __call__(self, *args, **kwargs):
        dct = {}
        for key, val in self.instance.items():
            if isinstance(val, np.ndarray):
                dct[key] = np.copy(val)
            else:
                dct[key] = val
        dct.update(self.__wrapped__(*args, **kwargs))
        return self.cls(**dct)

    def inplace_update(self, *args, **kwargs):
        dct = self.__wrapped__(*args, **kwargs)
        for key, val in dct.items():
            self.instance.__setattr__(key, val)

class DataContainer:
    """Abstract data container class.

    Parameters
    ----------
    **kwargs : dict
        Values of the attributes specified in `attr_set` and `init_set`.

    Attributes
    ----------
    attr_set : set
        Set of attributes in the container which are necessary
        to initialize in the constructor.
    init_set : set
        Set of optional data attributes.

    Raises
    ------
    ValueError
        If an attribute specified in `attr_set` has not been provided.
    """
    attr_set, init_set = {}, {}

    def __init__(self, **kwargs):
        self.__dict__['attr_dict'] = {}
        for attr in self.attr_set:
            if kwargs.get(attr) is None:
                raise ValueError('Attribute {:s} has not been provided'.format(attr))
            else:
                self.attr_dict[attr] = kwargs.get(attr)
        for attr in self.init_set:
            self.attr_dict[attr] = kwargs.get(attr)

    def __iter__(self):
        return self.attr_dict.__iter__()

    def __contains__(self, attr):
        return attr in self.attr_dict

    def __getattr__(self, attr):
        if attr in self:
            return self.attr_dict[attr]
        else:
            raise AttributeError(attr + " doesn't exist")

    def __repr__(self):
        return self.attr_dict.__repr__()

    def __setattr__(self, attr, value):
        if attr in self.__dict__['attr_dict']:
            self.__dict__['attr_dict'][attr] = value
        else:
            super(DataContainer, self).__setattr__(attr, value)

    def __str__(self):
        return self.attr_dict.__str__()

    def get(self, attr, value=None):
        """Retrieve a dataset, return `value` if the attribute is not found.

        Parameters
        ----------
        attr : str
            Data attribute.
        value : object, optional
            Data which is returned if the attribute is not found.

        Returns
        -------
        object
            Attribute's data stored in the container,
            `value` if `attr` is not found.
        """
        return self.attr_dict.get(attr, value)

    def keys(self):
        """Return the list of attributes stored in the container.

        Returns
        -------
        dict_keys
            List of attributes stored in the container.
        """
        return self.attr_dict.keys()

    def items(self):
        """Return (key, value) pairs of the datasets stored in the container.

        Returns
        -------
        dict_items
            (key, value) pairs of the datasets stored in the container.
        """
        return self.attr_dict.items()

    def values(self):
        """Return the attributes' data stored in the container.

        Returns
        -------
        dict_values
            List of data stored in the container.
        """
        return self.attr_dict.values()



"""Routines for model regression based on nonlinear
least-squares algorithm. :class:`pyrost.AbberationsFit` fit the
lens' abberations profile with the polynomial function
using nonlinear least-squares algorithm.

Examples
--------
Generate a :class:`pyrost.AbberationsFit` object from
:class:`pyrost.STData` container object `st_data` as follows:

>>> fit_obj = AbberationsFit.import_data(st_data)

Fit a pixel abberations profile with a third
order polynomial:

>>> fit = fit_obj.fit(max_order=3)
>>> print(fit)
{'alpha': -0.04718488324311934,
 'fit': array([-9.03305155e-04,  2.14699128e+00, -1.17287983e+03]),
 'ph_fit': array([-9.81298119e-07,  3.49854945e-03, -3.82244504e+00,  1.26179239e+03]),
 'rel_err': array([0.02331385, 0.01966198, 0.01679612]),
 'r_sq': 0.9923840802879347}
"""
import numpy as np
from scipy.optimize import least_squares
from .data_container import DataContainer, dict_to_object

class LeastSquares:
    """Basic nonlinear least-squares fit class.
    Based on :func:`scipy.optimize.least_squares`.

    See Also
    --------
    :func:`scipy.optimize.least_squares` : Full nonlinear least-squares
        algorithm description.
    """
    @classmethod
    def model(cls, fit, x, roi):
        """Return values of polynomial model function.

        Parameters
        ----------
        x : numpy.ndarray
            Array of x coordinates.

        Returns
        -------
        numpy.ndarray
            Array of polynomial function values.
        """
        return np.polyval(fit, x[roi[0]:roi[1]])

    @classmethod
    def errors(cls, fit, x, y, roi):
        """Return an array of model residuals.

        Parameters
        ----------
        fit : numpy.ndarray
            Array of fit coefficients.
        x : numpy.ndarray
            Array of x coordinates.
        y : numpy.ndarray
            Array of y coordinates.

        Returns
        -------
        numpy.ndarray
            Array of model residuals.
        """
        return cls.model(fit, x, roi) - y[roi[0]:roi[1]]

    @classmethod
    def fit(cls, x, y, max_order=2, xtol=1e-14, ftol=1e-14, loss='cauchy', roi=None):
        """Fit `x`, `y` with polynomial function using
        :func:`scipy.optimise.least_squares`.

        Parameters
        ----------
        x : numpy.ndarray
            Array of x coordinates.
        y : numpy.ndarray
            Array of y coordinates.
        max_order : int, optional
            Maximum order of the polynomial model function.
        xtol : float, optional
            Tolerance for termination by the change of the independent variables.
        ftol : float, optional
            Tolerance for termination by the change of the cost function.
        loss : {'linear', 'soft_l1', 'huber', 'cauchy', 'arctan'}, optional
            Determines the loss function. The following keyword values are
            allowed:

            * 'linear : ``rho(z) = z``. Gives a standard
              least-squares problem.
            * 'soft_l1' : ``rho(z) = 2 * ((1 + z)**0.5 - 1)``. The smooth
              approximation of l1 (absolute value) loss. Usually a good
              choice for robust least squares.
            * 'huber' : ``rho(z) = z if z <= 1 else 2*z**0.5 - 1``. Works
              similarly to 'soft_l1'.
            * 'cauchy'' (default) : ``rho(z) = ln(1 + z)``. Severely weakens
              outliers influence, but may cause difficulties in optimization
              process.
            * 'arctan' : ``rho(z) = arctan(z)``. Limits a maximum loss on
              a single residual, has properties similar to 'cauchy'.

        Returns
        -------
        fit : numpy.ndarray
            Array of fit coefficients.
        err : numpy.ndarray
            Vector of errors of the `fit` fit coefficients.
        """
        if roi is None:
            roi = (0, x.size)
        fit = least_squares(cls.errors, np.zeros(max_order + 1),
                            loss=loss, args=(x, y, roi), xtol=xtol, ftol=ftol)
        r_sq = 1 - np.sum(cls.errors(fit.x, x, y, roi)**2) / np.sum((y[roi[0]:roi[1]].mean() - y[roi[0]:roi[1]])**2)
        if np.linalg.det(fit.jac.T.dot(fit.jac)):
            cov = np.linalg.inv(fit.jac.T.dot(fit.jac))
            err = np.sqrt(np.sum(fit.fun**2) / (fit.fun.size - fit.x.size) * np.abs(np.diag(cov)))
        else:
            err = 0
        return fit.x, err, r_sq

class AbberationsFit(DataContainer):
    """Lens' abberations profile model regression using
    nonlinear least-squares algorithm. :class:`AbberationsFit`
    is capable of fitting lens' pixel abberations, deviation 
    angles, and phase profile with polynomial function.
    Based on :func:`scipy.optimise.least_squares`.

    Parameters
    ----------
    **kwargs : dict
        Dictionary of the attributes' data specified in `attr_set`
        and `init_set`.

    Attributes
    ----------
    attr_set : set
        Set of attributes in the container which are necessary
        to initialize in the constructor.

    Raises
    ------
    ValueError
        If an attribute specified in `attr_set` has not been provided.

    Notes
    -----
    Necessary attributes:

    * defocus : Defocus distance [m].
    * distance : Sample-to-detector distance [m].
    * phase : Abberations phase profile [rad].
    * pixels : Pixel coordinates [pixels].
    * pixel_abberations : Pixel abberations profile [pixels].
    * pixel_size : Pixel's size [m].
    * roi : Region of interest.
    * theta_abberations : 
    * wavelength : Incoming beam's wavelength [m].

    See Also
    --------
    :func:`scipy.optimize.least_squares` : Full nonlinear least-squares
        algorithm description.
    """
    attr_set = {'defocus', 'distance', 'phase', 'pixels', 'pixel_abberations',
                'pixel_size', 'wavelength'}
    init_set = {'roi', 'theta_abberations'}
    fs_lookup = {'defocus': 'defocus_fs', 'pixel_size': 'x_pixel_size'}
    ss_lookup = {'defocus': 'defocus_ss', 'pixel_size': 'y_pixel_size'}

    def __init__(self, **kwargs):
        super(AbberationsFit, self).__init__(**kwargs)
        self.pix_ap = self.pixel_size / self.distance
        if self.roi is None:
            self.roi = np.array([0, self.pixels.size])
        if self.theta_abberations is None:
            self.theta_abberations = self.pixel_abberations * self.pix_ap

    @classmethod
    def import_data(cls, st_data, center=0, axis=1):
        """Return a new :class:`AbberationsFit` object
        with all the necessary data attributes imported from
        the :class:`STData` container object `st_data`.

        Parameters
        ----------
        st_data : STData
            :class:`STData` container object.
        axis : int, optional
            Detector's axis (0 - slow axis, 1 - fast axis).

        Returns
        -------
        AbberationsFit
            A new :class:`AbberationsFit` object.
        """
        data_dict = {attr: st_data.get(attr) for attr in cls.attr_set}
        if axis == 0:
            data_dict.update({attr: st_data.get(data_attr) for attr, data_attr in cls.ss_lookup.items()})
        elif axis == 1:
            data_dict.update({attr: st_data.get(data_attr) for attr, data_attr in cls.fs_lookup.items()})
        else:
            raise ValueError('invalid axis value: {:d}'.format(axis))
        data_dict['pixel_abberations'] = data_dict['pixel_abberations'][axis].mean(axis=1 - axis)
        data_dict['pixels'] = np.arange(st_data.roi[2 * axis], st_data.roi[2 * axis + 1]) - center
        data_dict['phase'] = data_dict['phase'].mean(axis=1 - axis)
        return cls(**data_dict)

    @dict_to_object
    def crop_data(self, roi):
        """Return a new :class:`AbberationsFit` object with the updated `roi`.

        Parameters
        ----------
        roi : iterable
            Region of interest in the detector plane.

        Returns
        -------
        STData
            New :class:`AbberationsFit` object with the updated `roi`.
        """
        return {'roi': np.asarray(roi, dtype=int)}

    def get(self, attr, value=None):
        """Return a dataset with `mask` and `roi` applied.
        Return `value` if the attribute is not found.

        Parameters
        ----------
        attr : str
            Attribute to fetch.
        value : object, optional
            Return if `attr` is not found.

        Returns
        -------
        numpy.ndarray or object
            `attr` dataset with `mask` and `roi` applied.
            `value` if `attr` is not found.
        """
        if attr in self:
            val = super(AbberationsFit, self).get(attr)
            if not val is None:
                if attr in ['phase', 'pixels', 'pixel_abberations', 'theta_abberations']:
                    val = val[self.roi[0]:self.roi[1]]
            return val
        else:
            return value

    def model(self, fit):
        """Return the polynomial function values of
        lens' deviation angles fit.

        Parameters
        ----------
        fit : numpy.ndarray
            Lens` pixel abberations fit coefficients.
        roi : iterable, optional
            Region of interest. Full region if `roi` is None.

        Returns
        -------
        numpy.ndarray
            Array of polynomial function values.
        """
        return LeastSquares.model(fit, self.pixels, self.roi)

    def pix_to_phase(self, fit):
        """Convert fit coefficients from pixel
        abberations fit to abberations phase fit.

        Parameters
        ----------
        fit : numpy.ndarray
            Lens' pixel abberations fit coefficients.
        roi : iterable, optional
            Region of interest. Full region if `roi` is None.

        Returns
        -------
        numpy.ndarray
            Lens` phase abberations fit coefficients.
        """
        nfit = np.zeros(fit.size + 1)
        nfit[:-1] = 2 * np.pi * fit * self.pix_ap**2 * self.defocus / self.wavelength
        nfit[:-1] /= np.arange(1, fit.size + 1)[::-1]
        nfit[-1] = -self.model(nfit).mean()
        return nfit

    def phase_to_pix(self, ph_fit):
        """Convert fit coefficients from pixel
        abberations fit to abberations phase fit.

        Parameters
        ----------
        ph_fit : numpy.ndarray
            Lens` phase abberations fit coefficients.

        Returns
        -------
        numpy.ndarray
            Lens' pixel abberations fit coefficients.
        """
        fit = self.wavelength * ph_fit[:-1] / (2 * np.pi * self.pix_ap**2 * self.defocus)
        fit *= np.arange(1, ph_fit.size)[::-1]
        return fit

    def fit(self, max_order=2, xtol=1e-14, ftol=1e-14, loss='cauchy'):
        """Fit lens' pixel abberations with polynomial function using
        :func:`scipy.optimise.least_squares`.

        Parameters
        ----------
        max_order : int, optional
            Maximum order of the polynomial model function.
        xtol : float, optional
            Tolerance for termination by the change of the independent
            variables.
        ftol : float, optional
            Tolerance for termination by the change of the cost function.
        loss : {'linear', 'soft_l1', 'huber', 'cauchy', 'arctan'}, optional
            Determines the loss function. The following keyword values
            are allowed:

            * 'linear' : ``rho(z) = z``. Gives a standard
              least-squares problem.
            * 'soft_l1' : ``rho(z) = 2 * ((1 + z)**0.5 - 1)``. The smooth
              approximation of l1 (absolute value) loss. Usually a good
              choice for robust least squares.
            * 'huber' : ``rho(z) = z if z <= 1 else 2*z**0.5 - 1``. Works
              similarly to 'soft_l1'.
            * 'cauchy' (default) : ``rho(z) = ln(1 + z)``. Severely weakens
              outliers influence, but may cause difficulties in optimization
              process.
            * 'arctan' : ``rho(z) = arctan(z)``. Limits a maximum loss on
              a single residual, has properties similar to 'cauchy'.

        Returns
        -------
        dict
            :class:`dict` with the following fields defined:

            * alpha : Third order abberations ceofficient [rad/mrad^3].
            * fit : Array of the polynomial function coefficients of the
              pixel abberations fit.
            * ph_fit : Array of the polynomial function coefficients of
              the phase abberations fit.
            * rel_err : Vector of relative errors of the fit coefficients.
            * r_sq : ``R**2`` goodness of fit.

        See Also
        --------
        :func:`scipy.optimize.least_squares` : Full nonlinear least-squares
            algorithm description.
        """
        fit, err, r_sq = LeastSquares.fit(x=self.pixels, y=self.pixel_abberations, roi=self.roi,
                                          max_order=max_order, xtol=xtol, ftol=ftol, loss=loss)
        ph_fit = self.pix_to_phase(fit)
        if ph_fit.size >= 4:
            alpha = ph_fit[-4] / self.pix_ap**3 * 1e-9
        else:
            alpha = None
        return {'alpha': alpha, 'fit': fit, 'ph_fit': ph_fit, 'rel_err': np.abs(err / fit), 'r_sq': r_sq}

    def fit_phase(self, max_order=3, xtol=1e-14, ftol=1e-14, loss='linear'):
        """Fit lens' phase abberations with polynomial function using
        :func:`scipy.optimise.least_squares`.

        Parameters
        ----------
        max_order : int, optional
            Maximum order of the polynomial model function.
        xtol : float, optional
            Tolerance for termination by the change of the independent
            variables.
        ftol : float, optional
            Tolerance for termination by the change of the cost function.
        loss : {'linear', 'soft_l1', 'huber', 'cauchy', 'arctan'}, optional
            Determines the loss function. The following keyword values
            are allowed:

            * 'linear' : ``rho(z) = z``. Gives a standard
              least-squares problem.
            * 'soft_l1' : ``rho(z) = 2 * ((1 + z)**0.5 - 1)``. The smooth
              approximation of l1 (absolute value) loss. Usually a good
              choice for robust least squares.
            * 'huber' : ``rho(z) = z if z <= 1 else 2*z**0.5 - 1``. Works
              similarly to 'soft_l1'.
            * 'cauchy' (default) : ``rho(z) = ln(1 + z)``. Severely weakens
              outliers influence, but may cause difficulties in optimization
              process.
            * 'arctan' : ``rho(z) = arctan(z)``. Limits a maximum loss on
              a single residual, has properties similar to 'cauchy'.
        roi : iterable, optional
            Region of interest. Full region if `roi` is None.

        Returns
        -------
        dict
            :class:`dict` with the following fields defined:

            * alpha : Third order abberations ceofficient [rad/mrad^3].
            * fit : Array of the polynomial function coefficients of the
              pixel abberations fit.
            * ph_fit : Array of the polynomial function coefficients of
              the phase abberations fit.
            * rel_err : Vector of relative errors of the fit coefficients.
            * r_sq : ``R**2`` goodness of fit.

        See Also
        --------
        :func:`scipy.optimize.least_squares` : Full nonlinear least-squares
            algorithm description.
        """
        ph_fit, err, r_sq = LeastSquares.fit(x=self.pixels, y=self.phase, roi=self.roi,
                                             max_order=max_order, xtol=xtol, ftol=ftol, loss=loss)
        fit = self.phase_to_pix(ph_fit)
        if ph_fit.size >= 4:
            alpha = ph_fit[-4] / self.pix_ap**3 * 1e-9
        else:
            alpha = None
        return {'alpha': alpha, 'fit': fit, 'ph_fit': ph_fit, 'rel_err': np.abs(err / ph_fit)[:-1], 'r_sq': r_sq}



"""
Examples
--------
Generate the default CXI protocol.

>>> import pyrost as rst
>>> rst.cxi_protocol()
{'config': {'float_precision': 'float64'}, 'datatypes': {'basis_vectors':
'float', 'data': 'float', 'defocus': 'float', '...': '...'}, 'default_paths':
{'basis_vectors': '/entry_1/instrument_1/detector_1/basis_vectors',
'data': '/entry_1/data_1/data', 'defocus': '/speckle_tracking/defocus', '...': '...'}}

Or generate the default CXI loader.

>>> rst.loader()
{'config': {'float_precision': 'float64'}, 'datatypes': {'basis_vectors':
'float', 'data': 'float', 'defocus': 'float', '...': '...'}, 'default_paths':
{'basis_vectors': '/entry_1/instrument_1/detector_1/basis_vectors', 'data':
'/entry_1/data_1/data', 'defocus': '/speckle_tracking/defocus', '...': '...'},
'load_paths': {'good_frames': ['/speckle_tracking/good_frames',
'/frame_selector/good_frames', '/process_3/good_frames', '...'], 'mask':
['/speckle_tracking/mask', '/mask_maker/mask', '/entry_1/instrument_1/detector_1/mask',
'...'], 'translations': ['/entry_1/sample_1/geometry/translations',
'/entry_1/sample_1/geometry/translation', '/pos_refine/translation', '...'],
'...': '...'}, 'policy': {'basis_vectors': 'True', 'data': 'True', 'defocus':
'True', '...': '...'}}
"""
import os
import configparser
import re
import h5py
import numpy as np
from .data_processing import STData

ROOT_PATH = os.path.dirname(__file__)
PROTOCOL_FILE = os.path.join(ROOT_PATH, 'config/cxi_protocol.ini')

class hybridmethod:
    """Hybrid method descriptor supporting
    two distinct methodsbound to class and instance.

    Parameters
    ----------
    fclass : method
        Class bound method.
    finstance : method, optional
        Instance bound method.
    doc : str, optional
        Method's docstring.

    Attributes
    ----------
    fclass : method
        Class bound method.
    finstance : method
        Instance bound method.
    doc : str
        Method's dosctring.
    """
    def __init__(self, fclass, finstance=None, doc=None):
        self.fclass, self.finstance = fclass, finstance
        self.__doc__ = doc or fclass.__doc__
        self.__isabstractmethod__ = bool(getattr(fclass, '__isabstractmethod__', False))

    def classmethod(self, fclass):
        """Class method decorator

        Parameters
        ----------
        fclass : method
            Class bound method.

        Returns
        -------
        hybridmethod
            A new instance with the class bound method added to the object
        """
        return type(self)(fclass, self.finstance, None)

    def instancemethod(self, finstance):
        """Instance method decorator

        Parameters
        ----------
        finstance : method
            Instance bound method.

        Returns
        -------
        hybridmethod
            A new instance with the instance bound method added to the object
        """
        return type(self)(self.fclass, finstance, self.__doc__)

    def __get__(self, instance, cls):
        if instance is None or self.finstance is None:
            return self.fclass.__get__(cls, None)
        return self.finstance.__get__(instance, cls)

class INIParser:
    """INI files parser class with the methods for importing and exporting ini files
    and Python dictionaries.

    Parameters
    ----------
    **kwargs : dict
        Attributes specified in `attr_dict`.

    Attributes
    ----------
    err_txt : str
        Error text.
    known_types : dict
        Look-up dictionary of supported types for formatting.
    attr_dict : dict
        Dictionary of provided attributes.
    fmt_dict : dict
        Dictionary of attributes' types used for formatting.

    Raises
    ------
    AttributeError
        If the attribute specified in `attr_dict`
        has not been provided in keyword arguments ``**kwargs``.
    """
    err_txt = "Wrong format key '{0:s}' of option '{1:s}'"
    known_types = {'int': int, 'bool': bool, 'float': float, 'str': str}
    attr_dict, fmt_dict = {}, {}
    LIST_SPLITTER = r'\s*,\s*'
    LIST_MATCHER = r'^\[([\s\S]*)\]$'
    FMT_LEN = 3

    def __init__(self, **kwargs):
        for section in self.attr_dict:
            self.__dict__[section] = {}
            if 'ALL' in self.attr_dict[section]:
                if section in kwargs:
                    for option in kwargs[section]:
                        self._init_value(section, option, kwargs)
            else:
                for option in self.attr_dict[section]:
                    self._init_value(section, option, kwargs)

    def _init_value(self, section, option, kwargs):
        fmt = self.get_format(section, option)
        if option in kwargs[section]:
            if isinstance(kwargs[section][option], list):
                self.__dict__[section][option] = [fmt(part) for part in kwargs[section][option]]
            else:
                self.__dict__[section][option] = fmt(kwargs[section][option])
        else:
            raise AttributeError("The '{:s}' option has not been provided".format(option))

    @classmethod
    def read_ini(cls, protocol_file):
        """Read the `protocol_file` and return an instance of
        :class:`configparser.ConfigParser` class.

        Parameters
        ----------
        protocol_file : str
            Path to the file.

        Returns
        -------
        configparser.ConfigParser
            Parser object with all the data contained in the
            INI file.

        Raises
        ------
        ValueError
            If the file doesn't exist.
        """
        if not os.path.isfile(protocol_file):
            raise ValueError("File {:s} doesn't exist".format(protocol_file))
        ini_parser = configparser.ConfigParser()
        ini_parser.read(protocol_file)
        return ini_parser

    @classmethod
    def get_format(cls, section, option):
        """Return the attribute's format specified by `fmt_dict`.

        Parameters
        ----------
        section : str
            Attribute's section.

        option : str
            Attribute's name.

        Returns
        -------
        type
            Type of the attribute.
        """
        fmt = cls.fmt_dict.get(os.path.join(section, option))
        if not fmt:
            fmt = cls.fmt_dict.get(section)
        return cls.known_types[fmt]

    @classmethod
    def get_value(cls, ini_parser, section, option):
        """Return an attribute from an INI file's parser object `ini_parser`.

        Parameters
        ----------
        ini_parser : configparser.ConfigParser
            A parser object of an INI file.

        section : str
            Attribute's section.

        option : str
            Attribute's option.

        Returns
        -------
            Attribute's value imported from the INI file.
        """
        fmt = cls.get_format(section, option)
        string = ini_parser.get(section, option)
        is_list = re.search(cls.LIST_MATCHER, string)
        if is_list:
            return [fmt(part.strip())
                    for part in re.split(cls.LIST_SPLITTER, is_list.group(1))]
        else:
            return fmt(string.strip())

    @classmethod
    def _import_ini(cls, protocol_file):
        ini_parser = cls.read_ini(protocol_file)
        kwargs = {}
        for section in cls.attr_dict:
            kwargs[section] = {}
            if 'ALL' in cls.attr_dict[section]:
                if section in ini_parser:
                    for option in ini_parser[section]:
                        val = cls.get_value(ini_parser, section, option)
                        kwargs[section][option] = val
            else:
                for option in cls.attr_dict[section]:
                    val = cls.get_value(ini_parser, section, option)
                    kwargs[section][option] = val
        return kwargs

    @classmethod
    def import_ini(cls, protocol_file):
        """Initialize an :class:`INIParser` object class with an
        ini file.

        Parameters
        ----------
        protocol_file : str
            Path to the file.

        Returns
        -------
        INIParser
            An :class:`INIParser` object with all the attributes imported
            from the INI file.
        """
        return cls(**cls._import_ini(protocol_file))

    @classmethod
    def _format(cls, obj):
        crop_obj = {}
        for key, val in list(obj.items())[:cls.FMT_LEN]:
            if isinstance(val, dict):
                val = cls._format(val)
            elif isinstance(val, list):
                val = val[:cls.FMT_LEN] + ['...']
            crop_obj[key] = val
        if len(obj) > cls.FMT_LEN:
            crop_obj['...'] = '...'
        return crop_obj

    def __repr__(self):
        crop_dict = {key: self._format(val) for key, val in self.export_dict().items()}
        return crop_dict.__repr__()

    def __str__(self):
        crop_dict = {key: self._format(val) for key, val in self.export_dict().items()}
        return crop_dict.__str__()

    def export_dict(self):
        """Return a :class:`dict` object with all the attributes.

        Returns
        -------
        dict
            Dictionary with all the attributes conntained in the object.
        """
        return {section: self.__dict__[section] for section in self.attr_dict}

    @hybridmethod
    def export_ini(cls, **kwargs):
        """Return a :class:`configparser.ConfigParser` object
        with all the attributes exported the class.

        Parameters
        ----------
        **kwargs : dict
            Extra parameters to export to the
            :class:`configparser.ConfigParser` object.

        Returns
        ------
        configparser.ConfigParser
            A parser object with all the attributes contained in
            `attr_dict`.
        """
        ini_parser = configparser.ConfigParser()
        for section in cls.attr_dict:
            if 'ALL' in cls.attr_dict[section]:
                ini_parser[section] = kwargs[section]
            else:
                ini_parser[section] = {option: kwargs[section][option]
                                       for option in cls.attr_dict[section]}
        return ini_parser

    @export_ini.instancemethod
    def export_ini(self):
        """Return a :class:`configparser.ConfigParser` object
        with all the attributes exported from the object.

        Returns
        -------
        configparser.ConfigParser
            A parser object with all the attributes contained in
            the object.
        """
        return type(self).export_ini(**self.__dict__)

class Protocol(INIParser):
    """CXI protocol class. Contains a CXI file tree path with
    the paths written to all the data attributes necessary for
    the Speckle Tracking algorithm, corresponding attributes'
    data types, and floating point precision.

    Parameters
    ----------
    datatypes : dict, optional
        Dictionary with attributes' datatypes. 'float', 'int', or 'bool'
        are allowed.
    default_paths : dict, optional
        Dictionary with attributes' CXI default file paths.
    float_precision : {'float32', 'float64'}, optional
        Floating point precision.

    Attributes
    ----------
    config : dict
        Protocol configuration.
    datatypes : dict
        Dictionary with attributes' datatypes. 'float', 'int', or 'bool'
        are allowed.
    default_paths : dict
        Dictionary with attributes' CXI default file paths.

    See Also
    --------
    protocol : Full list of data attributes and configuration
        parameters.
    """
    attr_dict = {'config': ('float_precision', ), 'datatypes': ('ALL', ), 'default_paths': ('ALL', )}
    fmt_dict = {'config': 'str','datatypes': 'str', 'default_paths': 'str'}

    def __init__(self, datatypes=None, default_paths=None, float_precision='float64'):
        if datatypes is None:
            datatypes = {}
        if default_paths is None:
            default_paths = {}
        datatypes = {attr: val for attr, val in datatypes.items() if attr in default_paths}
        default_paths = {attr: val for attr, val in default_paths.items() if attr in datatypes}
        super(Protocol, self).__init__(config={'float_precision': float_precision},
                                       datatypes=datatypes, default_paths=default_paths)

        if self.config['float_precision'] == 'float32':
            self.known_types['float'] = np.float32
        elif self.config['float_precision'] == 'float64':
            self.known_types['float'] = np.float64
        else:
            raise ValueError('Invalid float precision: {:s}'.format(self.config['float_precision']))

    @classmethod
    def import_ini(cls, protocol_file):
        """Initialize an :class:`Protocol` object class with an
        ini file.

        Parameters
        ----------
        protocol_file : str
            Path to the file.

        Returns
        -------
        Protocol
            An :class:`Protocol` object with all the attributes imported
            from the INI file.
        """
        kwargs = cls._import_ini(protocol_file)
        return cls(datatypes=kwargs['datatypes'], default_paths=kwargs['default_paths'],
                   float_precision=kwargs['config']['float_precision'])

    def parser_from_template(self, path):
        """Return a :class:`configparser.ConfigParser` object using
        an ini file template.

        Parameters
        ----------
        path : str
            Path to the ini file template.

        Returns
        -------
        configparser.ConfigParser
            Parser object with the attributes populated according
            to the protocol.
        """
        ini_template = configparser.ConfigParser()
        ini_template.read(path)
        parser = configparser.ConfigParser()
        for section in ini_template:
            parser[section] = {option: ini_template[section][option].format(**self.default_paths)
                               for option in ini_template[section]}
        return parser

    def __iter__(self):
        return self.default_paths.__iter__()

    def __contains__(self, attr):
        return attr in self.default_paths

    def get_default_path(self, attr, value=None):
        """Return the atrribute's default path in the CXI file.
        Return `value` if `attr` is not found.

        Parameters
        ----------
        attr : str
            The attribute to look for.

        value : str, optional
            value which is returned if the `attr` is not found.

        Returns
        -------
        str or None
            Attribute's cxi file path.
        """
        return self.default_paths.get(attr, value)

    def get_dtype(self, attr, value=None):
        """Return the attribute's data type.
        Return `value` if `attr` is not found.

        Parameters
        ----------
        attr : str
            The data attribute.

        value : str, optional
            value which is returned if the `attr` is not found.

        Returns
        -------
        type or None
            Attribute's data type.
        """
        return self.known_types.get(self.datatypes.get(attr), value)

    def read_cxi(self, attr, cxi_file, cxi_path=None, dtype=None):
        """Read `attr` from the CXI file `cxi_file` at the path
        defined by the protocol. If `cxi_path` or `dtype` argument
        are provided, it will override the protocol.

        Parameters
        ----------
        attr : str
            Data attribute.
        cxi_file : h5py.File
            h5py File object of the CXI file.
        cxi_path : str, optional
            Path to the data attribute. If `cxi_path` is None,
            the path will be inferred according to the protocol.
        dtype : type, optional
            Data type of the attribute. If `dtype` is None,
            the type will be inferred according to the protocol.

        Returns
        -------
        numpy.ndarray or None
            The value of the attribute extracted from the CXI file.
        """
        if cxi_path is None:
            cxi_path = self.get_default_path(attr, cxi_path)
        if cxi_path in cxi_file:
            return cxi_file[cxi_path][...].astype(self.get_dtype(attr, dtype))
        else:
            return None

    def write_cxi(self, attr, data, cxi_file, overwrite=True, cxi_path=None, dtype=None):
        """Write data to the CXI file `cxi_file` under the path
        specified by the protocol. If `cxi_path` or `dtype` argument
        are provided, it will override the protocol.

        Parameters
        ----------
        attr : str
            Data attribute.
        data : numpy.ndarray
            Data which is bound to be saved.
        cxi_file : h5py.File
            :class:`h5py.File` object of the CXI file.
        overwrite : bool, optional
            Overwrite the content of `cxi_file` if it's True.
        cxi_path : str, optional
            Path to the data attribute. If `cxi_path` is None,
            the path will be inferred according to the protocol.
        dtype : type, optional
            Data type of the attribute. If `dtype` is None,
            the type will be inferred according to the protocol.

        Raises
        ------
        ValueError
            If `overwrite` is False and the data is already present
            at the given location in `cxi_file`.
        """
        if data is None:
            pass
        else:
            if cxi_path is None:
                cxi_path = self.get_default_path(attr, cxi_path)
            if cxi_path in cxi_file:
                if overwrite:
                    del cxi_file[cxi_path]
                else:
                    raise ValueError('{:s} is already present in {:s}'.format(attr,
                                                                              cxi_file.filename))
            data = np.asarray(data, dtype=self.get_dtype(attr, dtype))
            cxi_file.create_dataset(cxi_path, data=data)

def cxi_protocol(datatypes=None, default_paths=None, float_precision=None):
    """Return the default CXI porotocol.

    Parameters
    ----------
    datatypes : dict, optional
        Dictionary with attributes' datatypes. 'float', 'int', or 'bool'
        are allowed.
    default_paths : dict, optional
        Dictionary with attributes' CXI default file paths.
    float_precision : {'float32', 'float64'}, optional
        Floating point precision.

    Returns
    -------
    Protocol
        Default CXI protocol.

    See Also
    --------
    protocol : Full list of data attributes and configuration
        parameters.
    """
    if datatypes is None:
        datatypes = {}
    if default_paths is None:
        default_paths = {}
    kwargs = Protocol.import_ini(PROTOCOL_FILE).export_dict()
    kwargs['datatypes'].update(**datatypes)
    kwargs['default_paths'].update(**default_paths)
    if float_precision is None:
        float_precision = kwargs['config']['float_precision']
    return Protocol(datatypes=kwargs['datatypes'], default_paths=kwargs['default_paths'],
                    float_precision=float_precision)

class STLoader(Protocol):
    """Speckle Tracking data loader class. Loads data from a
    CXI file and returns a :class:`STData` container or a
    :class:`dict` with the data. Search data in the paths
    provided by `protocol` and `load_paths`.

    Parameters
    ----------
    protocol : Protocol
        Protocol object.
    load_paths : dict, optional
        Extra paths to the data attributes in a CXI file,
        which override `protocol`. Accepts only the attributes
        enlisted in `protocol`.
    policy : dict, optional
        A dictionary with loading policy. Contains all the
        attributes that are available in `protocol` and the
        corresponding flags. If a flag is True, the attribute
        will be loaded from a file. By default only the attributes
        necessary for a :class:`STData` container will be loaded.

    Attributes
    ----------
    config : dict
        Protocol configuration.
    datatypes : dict
        Dictionary with attributes' datatypes. 'float', 'int', or 'bool'
        are allowed.
    default_paths : dict
        Dictionary with attributes' CXI default file paths.
    load_paths : dict
        Extra set of paths to the attributes enlisted in `datatypes`.
    policy: dict
        Loading policy.

    See Also
    --------
    protocol : Full list of data attributes and configuration
        parameters.
    STData : Data container with all the data  necessary for
        Speckle Tracking.
    """
    attr_dict = {'config': ('float_precision', ), 'datatypes': ('ALL', ), 'default_paths': ('ALL', ),
                 'load_paths': ('ALL',), 'policy': ('ALL', )}
    fmt_dict = {'config': 'str','datatypes': 'str', 'default_paths': 'str', 'load_paths': 'str', 'policy': 'str'}

    def __init__(self, protocol, load_paths=None, policy=None):
        if load_paths is None:
            load_paths = {}
        else:
            load_paths = {attr: paths for attr, paths in load_paths.items()
                          if attr in protocol.default_paths}
        if policy is None:
            policy = {}
        else:
            policy = {attr: paths for attr, paths in policy.items()
                          if attr in protocol.default_paths}
        super(Protocol, self).__init__(config=protocol.config, datatypes=protocol.datatypes,
                                       default_paths=protocol.default_paths, load_paths=load_paths, policy=policy)

        if self.config['float_precision'] == 'float32':
            self.known_types['float'] = np.float32
        elif self.config['float_precision'] == 'float64':
            self.known_types['float'] = np.float64
        else:
            raise ValueError('Invalid float precision: {:s}'.format(self.config['float_precision']))

    @classmethod
    def import_ini(cls, protocol_file):
        """Initialize an :class:`STLoader` object class with an
        ini file.

        Parameters
        ----------
        protocol_file : str
            Path to the file.

        Returns
        -------
        STLoader
            An :class:`STLoader` object with all the attributes imported
            from the INI file.
        """
        kwargs = cls._import_ini(protocol_file)
        protocol = Protocol.import_ini(protocol_file)
        return cls(protocol=protocol, load_paths=kwargs['load_paths'], policy=kwargs['policy'])

    def get_load_paths(self, attr, value=None):
        """Return the atrribute's path in the cxi file.
        Return `value` if `attr` is not found.

        Parameters
        ----------
        attr : str
            The attribute to look for.

        value : str, optional
            value which is returned if the `attr` is not found.

        Returns
        -------
        list
            Set of attribute's paths.
        """
        paths = [super(STLoader, self).get_default_path(attr, value)]
        if attr in self.load_paths:
            paths.extend(self.load_paths[attr])
        return paths

    def get_policy(self, attr, value=None):
        policy = self.policy.get(attr, value)
        if isinstance(policy, str):
            return policy == 'True'
        else:
            return bool(policy)

    def get_protocol(self):
        """Return a CXI protocol from the loader.

        Returns
        -------
        Protocol
            CXI protocol.
        """
        return Protocol(datatypes=self.datatypes, default_paths=self.default_paths,
                        float_precision=self.config['float_precision'])

    def find_path(self, attr, cxi_file):
        """Find attribute's path in a CXI file `cxi_file`.

        Parameters
        ----------
        attr : str
            Data attribute.
        cxi_file : h5py.File
            :class:`h5py.File` object of the CXI file.

        Returns
        -------
        str or None
            Atrribute's path in the CXI file,
            return None if the attribute is not found.
        """
        paths = self.get_load_paths(attr)
        for path in paths:
            if path is None or path in cxi_file:
                return path
        else:
            return None

    def load_dict(self, path, **kwargs):
        """Load a CXI file and return a :class:`dict` with
        all the data fetched from the file.

        Parameters
        ----------
        path : str
            Path to the cxi file.
        **kwargs : dict, optional
            Dictionary of attribute values,
            which will be parsed to the `STData` object instead.

        Returns
        -------
        dict
            Dictionary with all the data fetched from the CXI file.
        """
        data_dict = {}
        with h5py.File(path, 'r') as cxi_file:
            for attr in self:
                if self.get_policy(attr, False):
                    cxi_path = self.find_path(attr, cxi_file)
                    if attr in kwargs and not kwargs[attr] is None:
                        data_dict[attr] = np.asarray(kwargs[attr], dtype=self.get_dtype(attr))
                    else:
                        data_dict[attr] = self.read_cxi(attr, cxi_file, cxi_path=cxi_path)
                else:
                    data_dict[attr] = None
        if not data_dict['defocus'] is None:
            data_dict['defocus_ss'] = data_dict['defocus']
            data_dict['defocus_fs'] = data_dict['defocus']
        return data_dict

    def load(self, path, **kwargs):
        """Load a CXI file and return an :class:`STData` class object.

        Parameters
        ----------
        path : str
            Path to the cxi file.
        **kwargs : dict
            Dictionary of attribute values,
            which will be parsed to the `STData` object instead.

        Returns
        -------
        STData
            Data container object with all the necessary data
            for the Speckle Tracking algorithm.
        """
        return STData(self.get_protocol(), **self.load_dict(path, **kwargs))

def loader(protocol=None, load_paths=None, policy=None):
    """Return the default CXI loader.

    Parameters
    ----------
    float_precision : {'float32', 'float64'}, optional
        Floating point precision.
    load_paths : dict, optional
        Extra paths to the data attributes in a CXI file,
        which override the :func:`cxi_protocol`. Accepts
        only the attributes enlisted in :func:`cxi_protocol`.
    policy : dict, optional
        A dictionary with loading policy. Contains all the
        attributes that are available in :func:`cxi_protocol`
        and the corresponding flags. If a flag is True, the
        attribute will be loaded from a file. By default only
        the attributes necessary for a :class:`STData` container
        will be loaded.

    Returns
    -------
    STLoader
        Default CXI loader.

    See Also
    --------
    cxi_protocol : Default CXI protocol.
    STLoader : Full loader class description.
    STData : Data container with all the data  necessary for
        Speckle Tracking.
    """
    kwargs = STLoader.import_ini(PROTOCOL_FILE).export_dict()
    if protocol is None:
        protocol = Protocol(datatypes=kwargs['datatypes'],
                            default_paths=kwargs['default_paths'],
                            float_precision=kwargs['config']['float_precision'])
    if load_paths is None:
        load_paths = {}
    if policy is None:
        policy = {}
    kwargs['load_paths'].update(**load_paths)
    kwargs['policy'].update(**policy)
    return STLoader(protocol, load_paths=kwargs['load_paths'],
                    policy=kwargs['policy'])



import os
import shutil
from datetime import datetime
import pytest
import pyrost as rst
import pyrost.simulation as st_sim

@pytest.fixture(params=[{'det_dist': 5e5, 'n_frames': 10, 'ap_x': 4,
                         'ap_y': 1, 'focus': 3e3, 'defocus': 2e2},
                        {'det_dist': 4.5e5, 'n_frames': 5, 'ap_x': 3,
                         'ap_y': 1.5, 'focus': 2e3, 'defocus': 1e2}],
                scope='session')
def st_params(request):
    """Return a default instance of simulation parameters.
    """
    return st_sim.parameters(**request.param)

@pytest.fixture(scope='session')
def sim_obj(st_params):
    sim_obj = st_sim.STSim(st_params)
    return sim_obj

@pytest.fixture(scope='session')
def ptych(sim_obj):
    data = sim_obj.ptychograph()
    return data

@pytest.fixture(params=['float32', 'float64'])
def loader(request):
    """
    Return the default loader.
    """
    protocol = rst.cxi_protocol(float_precision=request.param)
    return rst.loader(protocol=protocol)

@pytest.fixture(params=['float32', 'float64'])
def converter(request):
    """
    Return the default loader.
    """
    return st_sim.converter(float_precision=request.param)

@pytest.fixture(scope='session')
def temp_dir():
    now = datetime.now()
    path = now.strftime("temp_%m_%d_%H%M%S")
    os.mkdir(path)
    yield path
    shutil.rmtree(path)

@pytest.fixture(scope='function')
def ini_path(temp_dir):
    """Return a path to the experimental speckle tracking data.
    """
    path = os.path.join(temp_dir, 'test.ini')
    yield path
    os.remove(path)

@pytest.mark.st_sim
def test_st_params(st_params, ini_path):
    assert not os.path.isfile(ini_path)
    ini_parser = st_params.export_ini()
    with open(ini_path, 'w') as ini_file:
        ini_parser.write(ini_file)
    new_params = st_sim.STParams.import_ini(ini_path)
    assert new_params.export_dict() == st_params.export_dict()

@pytest.mark.st_sim
def test_ptych(ptych, st_params):
    assert len(ptych.shape) == 3
    assert ptych.shape[0] == st_params.n_frames

@pytest.mark.rst
def test_save_and_load_sim(converter, loader, ptych, sim_obj, temp_dir):
    assert os.path.isdir(temp_dir)
    converter.save_sim(ptych, sim_obj, temp_dir)
    cxi_path = os.path.join(temp_dir, 'data.cxi')
    assert os.path.isfile(cxi_path)
    data_dict = loader.load_dict(cxi_path)
    for attr in rst.STData.attr_set:
        assert not data_dict[attr] is None

@pytest.mark.rst
def test_st_update_sim(converter, ptych , sim_obj):
    st_data = converter.export_data(ptych, sim_obj)
    assert st_data.data.dtype == converter.protocol.known_types['float']
    st_obj = st_data.get_st()
    pixel_map0 = st_obj.pixel_map.copy()
    st_obj.iter_update(sw_fs=10, ls_pm=2.5, ls_ri=15,
                       verbose=True, n_iter=10)
    assert (st_obj.pixel_map == pixel_map0).all()
    assert st_obj.pixel_map.dtype == converter.protocol.known_types['float']

@pytest.mark.rst
def test_load_exp(path, roi, defocus, loader):
    assert os.path.isfile(path)
    data_dict = loader.load_dict(path=path, roi=roi, defocus=defocus)
    for attr in rst.STData.attr_set:
        assert not data_dict[attr] is None

@pytest.mark.rst
def test_st_update_exp(path, roi, defocus, loader):
    assert os.path.isfile(path)
    data = loader.load(path=path, roi=roi, defocus=defocus)
    assert data.data.dtype == loader.known_types['float']
    st_obj = data.get_st()
    pixel_map0 = st_obj.pixel_map.copy()
    st_obj.iter_update_gd(sw_fs=10, ls_pm=2.5, ls_ri=30,
                          verbose=True, n_iter=10)
    assert (st_obj.pixel_map == pixel_map0).all()
    assert st_obj.pixel_map.dtype == loader.known_types['float']

@pytest.mark.standalone
def test_full(converter, ptych, sim_obj):
    data = converter.export_data(ptych, sim_obj)
    assert data.data.dtype == converter.protocol.known_types['float']
    st_obj = data.get_st()
    st_res = st_obj.iter_update(sw_fs=10, ls_pm=2.5, ls_ri=15,
                                verbose=True, n_iter=10)
    data = data.update_phase(st_res)
    fit = data.fit_phase(axis=1)
    assert (st_obj.pixel_map != st_res.pixel_map).any()
    assert st_res.pixel_map.dtype == converter.protocol.known_types['float']
    assert not fit is None



"""
Examples
--------

Generate parameters, which could be later parsed to :class:`pyrost.simulation.STSim`
in order to perform the simulation.

>>> import pyrost.simulation as st_sim
>>> st_params = st_sim.parameters()
>>> print(st_params)
{'defocus': 400.0, 'det_dist': 2000000.0, 'step_size': 0.1, 'n_frames': 300,
'fs_size': 2000, 'ss_size': 1000, 'pix_size': 55.0, '...': '...'}
"""
import os
import numpy as np
from ..protocol import INIParser, ROOT_PATH
from ..bin import bar_positions, barcode_profile

PARAMETERS_FILE = os.path.join(ROOT_PATH, 'config/parameters.ini')

class STParams(INIParser):
    """Container with the experimental parameters of
    one-dimensional Speckle Tracking scan. All the experimental
    parameters are enlisted in `attr_dict`.

    Parameters
    ----------
    **kwargs : dict
        Values for the exerimental parameters specified
        in `attr_dict`.

    Attributes
    ----------
    attr_dict : dict
        Dictionary which contains all the experimental
        parameters.
    fmt_dict: dict
        Dictionary which specifies the data types of the
        parameters in `attr_dict`.

    See Also
    --------
    st_sim_param : Full list of experimental parameters.
    """
    attr_dict = {'exp_geom': ('defocus', 'det_dist', 'n_frames',
                              'step_size', 'step_rnd'),
                 'detector': ('fs_size', 'pix_size', 'ss_size'),
                 'source':   ('p0', 'th_s', 'wl'),
                 'lens':     ('alpha', 'ap_x', 'ap_y', 'focus', 'ab_cnt'),
                 'barcode':  ('bar_atn', 'bar_rnd', 'bar_sigma', 'bar_size',
                              'bulk_atn', 'offset'),
                 'system':   ('verbose',)}

    fmt_dict = {'exp_geom': 'float', 'exp_geom/n_frames': 'int',
                'detector': 'int', 'detector/pix_size': 'float',
                'source': 'float', 'lens': 'float', 'barcode': 'float',
                'system/verbose': 'bool'}
    FMT_LEN = 7

    @classmethod
    def lookup_dict(cls):
        """Look-up table between the sections and the parameters.

        Returns
        -------
        dict
            Look-up dictionary.
        """
        lookup = {}
        for section in cls.attr_dict:
            for option in cls.attr_dict[section]:
                lookup[option] = section
        return lookup

    def __init__(self, **kwargs):
        super(STParams, self).__init__(**kwargs)
        self.__dict__['_lookup'] = self.lookup_dict()

    def __iter__(self):
        return self._lookup.__iter__()

    def __contains__(self, attr):
        return attr in self._lookup

    def __getattr__(self, attr):
        if attr in self._lookup:
            return self.__dict__[self._lookup[attr]][attr]
        else:
            raise AttributeError(attr + " doesn't exist")

    def __setattr__(self, attr, value):
        if attr in self._lookup:
            fmt = self.get_format(self._lookup[attr], attr)
            self.__dict__[self._lookup[attr]][attr] = fmt(value)
        else:
            raise AttributeError(attr + ' not allowed')

    def __repr__(self):
        return self._format(self.export_dict()).__repr__()

    def __str__(self):
        return self._format(self.export_dict()).__str__()

    @classmethod
    def import_dict(cls, **kwargs):
        """Initialize experimental parameters from a dictionary `kwargs`.

        Parameters
        ----------
        **kwargs : dict
            Dictionary with experimental parameters.

        Returns
        -------
        STParams
            An :class:`STParams` object with the parameters from `kwargs`.
        """
        init_dict = {}
        for section in cls.attr_dict:
            init_dict[section] = {}
            for option in cls.attr_dict[section]:
                init_dict[section][option] = kwargs[option]
        return cls(**init_dict)

    def x_wavefront_size(self):
        r"""Return wavefront array size along the x axis, that
        satisfies the sampling condition.

        Returns
        -------
        n_x : int
            Array size.

        Notes
        -----
        The sampling condition when we propagate the wavefield from
        the lens plane to the sample plane is as follows:

        .. math::
            N_x >= \frac{\mathrm{max} \left( \Delta x_{lens}, \Delta x_{sample} \right)
            (\Delta x_{lens} + \Delta x_{sample})}{\lambda d} =
            \frac{4 a_x^2 \mathrm{max} \left( f, df \right)}{f^2 \lambda}

        Whereupon, the sampling condition when we propagate the wavefield to the detector
        plane is:

        .. math::
            N_x >= \frac{\Delta x_{sample} \Delta x_{det}}{\lambda d} =
            \frac{2 a_x n_{fs} \Delta_{pix} df}{f \lambda d_{det}}
        """
        nx_ltos = int(4 * self.ap_x**2 * max(self.focus, self.defocus) / self.focus**2 / self.wl)
        nx_stod = int(2 * self.fs_size * self.pix_size * self.ap_x * self.defocus / self.focus / self.wl / self.det_dist)
        return max(nx_ltos, nx_stod, self.fs_size)

    def y_wavefront_size(self):
        r"""Return wavefront array size along the y axis, that
        satisfies the sampling condition.

        Returns
        -------
        n_y : int
            Array size.

        Notes
        -----
        The sampling condition when we propagate the wavefield from
        the lens plane to the sample plane is as follows:

        .. math::
            N_y >= \frac{\mathrm{max} \left( \Delta y_{lens}, \Delta y_{sample} \right)
            (\Delta y_{lens} + \Delta y_{sample})}{\lambda d} =
            \frac{8 a_y^2}{(f + df) \lambda}

        Whereupon, the sampling condition when we propagate the wavefield to the detector
        plane is:

        .. math::
            N_y >= \frac{\Delta y_{sample} \Delta y_{det}}{\lambda d} =
            \frac{2 a_y n_{ss} \Delta_{pix}}{\lambda d_{det}}
        """
        ny_ltos = int(8 * self.ap_y**2 / (self.focus + self.defocus) / self.wl)
        ny_stod = int(2 * self.ss_size * self.pix_size * self.ap_y / self.wl / self.det_dist)
        return max(ny_ltos, ny_stod, self.ss_size)

    def lens_wavefronts(self, n_x=None, n_y=None, return_dxdy=False):
        r"""Return wavefields at the lens plane along x and y axes.

        Parameters
        ----------
        n_x : int, optional
            Array size along the x axis. Equals to
            :func:`STParams.x_wavefront_size` if it's None.
        n_y : int, optional
            Array size along the y axis. Equals to
            :func:`STParams.y_wavefront_size` if it's None.
        return_dxdy: bool, optional
            Return step sizes along x and y axes if it's True.

        Returns
        -------
        u0_x : numpy.ndarray
            Wavefront along the x axis.
        u0_y : numpy.ndarray
            Wavefront along the y axis.
        dx : float
            Step size along the x axis [um]. Only if
            `return_dxdy` is True.
        dy : float
            Step size along the y axis [um]. Only if
            `return_dxdy` is True.

        Notes
        -----
        The exit-surface at the lens plane:

        .. math::
            U_0(x) = \Pi(a_x x) \exp
            \left[ -\frac{j \pi x^2}{\lambda f} + j \alpha
            \left( \frac{x - x_{ab_cnt}}{f} \right)^3 \right]

        .. math::
            U_0(y) = \Pi(a_y y)
        """
        if n_x is None:
            n_x = self.x_wavefront_size()
        if n_y is None:
            n_y = self.y_wavefront_size()
        dx, dy = 2 * self.ap_x / n_x, 2 * self.ap_y / n_y
        x0 = dx * np.arange(-n_x // 2, n_x // 2)
        y0 = dy * np.arange(-n_y // 2, n_y // 2)
        x_cnt = (self.ab_cnt - 0.5) * self.ap_x
        u0_x = np.exp(1j * np.pi * x0**2 / self.wl / self.focus - \
                      1j * 1e9 * self.alpha * ((x0 - x_cnt) / self.focus)**3)
        u0_y = np.ones(n_y, dtype=np.complex128)
        u0_x[np.abs(x0) > self.ap_x / 2] = 0
        u0_y[np.abs(y0) > self.ap_y / 2] = 0
        if return_dxdy:
            return u0_x, u0_y, dx, dy
        else:
            return u0_x, u0_y

    def beam_span(self, dist):
        """Return beam span along the x axis at distance `dist`
        from the focal plane.

        Parameters
        ----------
        dist : float
            Distance from the focal plane [um].

        Returns
        -------
        th_lb : float
            Beam's lower bound [um].
        th_ub : float
            Beam's upper bound [um].
        """
        th_lb = -0.5 * self.ap_x / self.focus + self.wl / np.pi * self.alpha * \
                3.75e8 * (self.ap_x / self.focus)**2 / dist
        th_ub = 0.5 * self.ap_x / self.focus + self.wl / np.pi * self.alpha * \
                3.75e8 * (self.ap_x / self.focus)**2 / dist
        return np.tan(th_lb) * dist, np.tan(th_ub) * dist

    def bar_positions(self, dist):
        """Generate a coordinate array of barcode's bar positions at
        distance `dist` from focal plane.

        Parameters
        ----------
        dist : float
            Distance from the focal plane [um].

        Returns
        -------
        bar_pos : numpy.ndarray
            Array of barcode's bar coordinates.

        See Also
        --------
        bin.bar_positions : Full details of randomized barcode steps
            generation algorithm.
        """
        x0, x1 = self.beam_span(dist)
        return bar_positions(x0=x0 + self.offset, b_dx=self.bar_size, rd=self.bar_rnd,
                             x1=x1 + self.step_size * self.n_frames - self.offset)

    def sample_positions(self):
        """Generate an array of sample's translations with random deviation.

        Returns
        -------
        smp_pos : numpy.ndarray
            Array of sample translations [um].
        """
        rnd_arr = 2 * self.step_rnd * (np.random.random(self.n_frames) - 0.5)
        return self.step_size * (np.arange(self.n_frames) + rnd_arr)

    def barcode_profile(self, bar_pos, x_arr):
        """Generate a barcode's transmission profile at `x_arr`
        coordinates.

        Parameters
        ----------
        bar_pos : numpy.ndarray
            Array of barcode's bar positions [um].
        x_arr : numpy.ndarray
            Array of the coordinates, where the transmission
            coefficients are calculated [um].

        Returns
        -------
        b_tr : numpy.ndarray
            Barcode's transmission profile.

        See Also
        --------
        bin.barcode_profile : Full details of barcode's transmission
            profile generation algorithm.
        """
        return barcode_profile(bar_pos=bar_pos, x_arr=x_arr.ravel(), b_atn=self.bar_atn,
                               b_sgm=self.bar_sigma, blk_atn=self.bulk_atn).reshape(x_arr.shape)

    def source_curve(self, dist, dx):
        """Return source's rocking curve profile at `dist` distance from
        the lens.

        Parameters
        ----------
        dist : float
            Distance from the lens to the rocking curve profile [um].
        dx : float
            Sampling distance [um].

        Returns
        -------
        numpy.ndarray
            Source's rocking curve profile.
        """
        sc_sgm, n_sc = self.th_s * dist, np.ceil(8 * self.th_s * dist / dx)
        sc_x = dx * np.arange(-n_sc // 2, n_sc // 2 + 1)
        sc = np.exp(-sc_x**2 / 2 / sc_sgm**2)
        return sc / sc.sum()

    def fs_roi(self):
        """Return detector's region of interest along the fast axis.

        Returns
        -------
        fs_lb : int
            Beam's lower bound along the detector's fast axis.
        fs_ub : int
            Beam's upper bound along the detector's fast axis.
        """
        x_lb, x_ub = self.beam_span(self.det_dist)
        fs_lb = int(x_lb // self.pix_size) + self.fs_size // 2
        fs_ub = int(x_ub // self.pix_size) + self.fs_size // 2
        fs_lb = fs_lb if fs_lb > 0 else 0
        fs_ub = fs_ub if fs_ub < self.fs_size else self.fs_size
        return fs_lb, fs_ub

    def export_dict(self):
        """Export experimental parameters to :class:`dict`.

        Returns
        -------
        param_dict : dict
            Experimental parameters.
        """
        param_dict = {}
        for section in self.attr_dict:
            for option in self.attr_dict[section]:
                param_dict[option] = self.__dict__[section][option]
        return param_dict

def parameters(**kwargs):
    """Return the default :class:`STParams` object. Override any
    experimental parameters with `**kwargs`.

    Parameters
    ----------
    **kwargs : dict
        Dictionary which contains experimental
        parameters values.

    Returns
    -------
    st_params : STParams
        Default experimental parameters.

    See Also
    --------
    STParams : Full list of the experimental parameters.
    """
    st_params = STParams.import_ini(PARAMETERS_FILE).export_dict()
    st_params.update(**kwargs)
    return STParams.import_dict(**st_params)



"""Robust Speckle Tracking data processing algorithm.
:class:`pyrost.STData` contains all the necessary data for the Speckle
Tracking algorithm, and handy data processing tools to work
with the data. :class:`pyrost.SpeckleTracking` performs the main
Robust Speckle Tracking algorithm and yields reference image
and pixel mapping. :class:`pyrost.AbberationsFit` fit the lens'
abberations profile with the polynomial function using
nonlinear least-squares algorithm.

Examples
--------
Extract all the necessary data using a :func:`pyrost.loader` function.

>>> import pyrost as rst
>>> loader = rst.loader()
>>> rst_data = loader.load('results/test/data.cxi')

Perform the Robust Speckle Tracking using a :class:`pyrost.SpeckleTracking` object.

>>> rst_obj = rst_data.get_st()
>>> rst_res, errors = rst_obj.iter_update(sw_fs=150, ls_pm=3., ls_ri=10.,
...                                       verbose=True, n_iter=10)
Iteration No. 0: Total MSE = 0.150
Iteration No. 1: Total MSE = 0.077
Iteration No. 2: Total MSE = 0.052
Iteration No. 3: Total MSE = 0.050

Extract lens' abberations wavefront and fit it with a polynomial.

>>> rst_data.update_phase(rst_res)
>>> fit_res = rst_data.fit_phase()
>>> fit_res['ph_fit']
array([-5.19025587e+07, -8.63773622e+05,  3.42849675e+03,  2.98523995e+01,
        1.19773905e-02])
"""
from __future__ import division

import numpy as np
from scipy.ndimage import gaussian_filter, gaussian_gradient_magnitude
from .abberations_fit import AbberationsFit
from .data_container import DataContainer, dict_to_object
from .bin import make_whitefield, make_reference, update_pixel_map_gs, update_pixel_map_nm
from .bin import update_translations_gs, mse_frame, mse_total, ct_integrate

class STData(DataContainer):
    """Speckle Tracking algorithm data container class.
    Contains all the necessary data for the Robust Speckle
    Tracking algorithm (specified in `attr_set` and `init_set`),
    the list of all the :class:`SpeckleTracking` objects derived
    from it, and two :class:`AbberationsFit` objects to fit the phase
    along the fast and slow detector axes.

    Parameters
    ----------
    protocol : Protocol
        CXI :class:`Protocol` object.

    **kwargs : dict
        Dictionary of the attributes' data specified in `attr_set`
        and `init_set`.

    Attributes
    ----------
    attr_set : set
        Set of attributes in the container which are necessary
        to initialize in the constructor.
    init_set : set
        Set of optional data attributes.

    Raises
    ------
    ValueError
        If an attribute specified in `attr_set` has not been provided.

    Notes
    -----
    Necessary attributes:

    * basis_vectors : Detector basis vectors
    * data : Measured intensity frames.
    * distance : Sample-to-detector distance [m].
    * translations : Sample's translations [m].
    * wavelength : Incoming beam's wavelength [m].
    * x_pixel_size : Pixel's size along the fast detector axis [m].
    * y_pixel_size : Pixel's size along the slow detector axis [m].

    Optional attributes:

    * defocus_ss : Defocus distance for the slow detector axis [m].
    * defocus_fs : Defocus distance for the fast detector axis [m].
    * error_frame : MSE (mean-squared-error) of the reference image
      and pixel mapping fit per pixel.
    * good_frames : An array of good frames' indices.
    * mask : Bad pixels mask.
    * phase : Phase profile of lens' abberations.
    * pixel_map : The pixel mapping between the data at the detector's
      plane and the reference image at the reference plane.
    * pixel_abberations : Lens' abberations along the fast and
      slow axes in pixels.
    * pixel_translations : Sample's translations in the detector's
      plane in pixels.
    * reference_image : The unabberated reference image of the sample.
    * roi : Region of interest in the detector plane.
    * whitefield : Measured frames' whitefield.
    """
    attr_set = {'basis_vectors', 'data', 'distance', 'translations', 'wavelength',
                'x_pixel_size', 'y_pixel_size'}
    init_set = {'defocus_fs', 'defocus_ss', 'error_frame', 'good_frames', 'mask',
                'phase', 'pixel_abberations', 'pixel_map', 'pixel_translations',
                'reference_image', 'roi', 'whitefield'}

    def __init__(self, protocol, **kwargs):
        # Initialize protocol for the proper data type conversion in __setattr__
        self.__dict__['protocol'] = protocol

        # Initialize attr_dict
        super(STData, self).__init__(**kwargs)

        # Add protocol to the attr_dict in order to get the dict_update decorator working
        self.__dict__['attr_dict']['protocol'] = protocol

        # Initialize init_set attributes
        self._init_dict()

    def _init_dict(self):
        # Set ROI, good frames array, mask, and whitefield
        if self.roi is None:
            self.roi = np.array([0, self.data.shape[1], 0, self.data.shape[2]])
        if self.good_frames is None:
            self.good_frames = np.arange(self.data.shape[0])
        if self.mask is None or self.mask.shape != self.data.shape:
            self.mask = np.ones(self.data.shape)
        if self.whitefield is None:
            self.whitefield = make_whitefield(data=self.data[self.good_frames],
                                              mask=self.mask[self.good_frames])

        # Set a pixel map, deviation angles, and phase
        if not self.pixel_map is None:
            self.pixel_abberations = self.pixel_map
        self.pixel_map = np.indices(self.whitefield.shape)

        if self._isdefocus:
            # Set a pixel translations
            if self.pixel_translations is None:
                self.pixel_translations = (self.translations[:, None] * self.basis_vectors).sum(axis=-1)
                defocus = np.array([self.defocus_ss, self.defocus_fs])
                self.pixel_translations /= (self.basis_vectors**2).sum(axis=-1) * defocus / self.distance
                self.pixel_translations -= self.pixel_translations.mean(axis=0)

        # Initialize a list of SpeckleTracking objects
        self._st_objects = []

        # Initialize a list of AbberationsFit objects
        self._ab_fits = []
        if self._isphase:
            self._ab_fits.extend([AbberationsFit.import_data(self, axis=0),
                                  AbberationsFit.import_data(self, axis=1)])

    @property
    def _isdefocus(self):
        return not self.defocus_ss is None and not self.defocus_fs is None

    @property
    def _isphase(self):
        return not self.pixel_abberations is None and not self.phase is None

    def __setattr__(self, attr, value):
        if attr in self.attr_set | self.init_set:
            value = np.array(value, dtype=self.protocol.get_dtype(attr))
            super(STData, self).__setattr__(attr, value)
        else:
            super(STData, self).__setattr__(attr, value)

    @dict_to_object
    def crop_data(self, roi):
        """Return a new :class:`STData` object with the updated `roi`.

        Parameters
        ----------
        roi : iterable
            Region of interest in the detector plane.

        Returns
        -------
        STData
            New :class:`STData` object with the updated `roi`.
        """
        return {'roi': np.asarray(roi, dtype=int)}

    @dict_to_object
    def integrate_data(self, axis=1):
        """Return a new :class:`STData` object with the `data` summed
        over the `axis`.

        Parameters
        ----------
        axis : int
            Axis along which a sum is performed. 

        Returns
        -------
        STData
            New :class:`STData` object with the updated `data`,
            `whitefield`, `mask`, and `roi`.
        """
        roi = self.roi.copy()
        roi[2 * (axis - 1):2 * axis] = np.arange(2)
        return {'data': np.sum(self.data, axis=axis, keepdims=True), 'whitefield': None,
                'mask': None, 'roi': roi}

    @dict_to_object
    def mask_frames(self, good_frames):
        """Return a new :class:`STData` object with the updated
        good frames mask.

        Parameters
        ----------
        good_frames : iterable
            List of good frames' indices.

        Returns
        -------
        STData
            New :class:`STData` object with the updated `good_frames`
            and `whitefield`.
        """
        return {'good_frames': np.asarray(good_frames, dtype=np.int),
                'whitefield': None}

    @dict_to_object
    def update_mask(self, method='no-bad', pmin = 0., pmax=99.99, vmin=0, vmax=65535,
                    update='reset'):
        """Return a new :class:`STData` object with the updated
        bad pixels mask.

        Parameters
        ----------
        method : {'no-bad', 'eiger-bad', 'perc-bad'}, optional
            Bad pixels masking methods:

            * 'no-bad' (default) : No bad pixels.
            * 'range-bad' : Mask the pixels which values lie outside
              of (`vmin`, `vmax`) range.
            * 'perc-bad' : Mask the pixels which values lie outside
              of the (`pmin`, `pmax`) percentiles.

        vmin, vmax : float, optional
            Lower and upper intensity values of 'range-bad' masking
            method.

        pmin, pmax : float, optional
            Lower and upper percentage values of 'perc-bad' masking
            method.

        Returns
        -------
        STData
            New :class:`STData` object with the updated `mask`.
        """
        if method == 'no-bad':
            mask = np.ones(self.data.shape)
        elif method == 'range-bad':
            mask = (self.data > vmin) & (self.data < vmax)
        elif method == 'perc-bad':
            offsets = (self.data - np.median(self.data))
            mask = (offsets > np.percentile(offsets, pmin)) & \
                   (offsets < np.percentile(offsets, pmax))
        else:
            ValueError('invalid method argument')
        if update == 'reset':
            return {'mask': mask, 'whitefield': None}
        if update == 'multiply':
            return {'mask': self.mask * mask, 'whitefield': None}

    @dict_to_object
    def make_whitefield(self):
        """Return a new :class:`STData` object with the updated `whitefield`.

        Returns
        -------
        STData
            New :class:`STData` object with the updated `whitefield`.
        """
        return {'whitefield': None}

    @dict_to_object
    def make_pixel_map(self):
        """Return a new :class:`STData` object with the default `pixel_map`.

        Returns
        -------
        STData
            New :class:`STData` object with the updated `pixel_map`.
        """
        return {'pixel_map': None}

    @dict_to_object
    def update_defocus(self, defocus_fs, defocus_ss=None):
        """Return a new :class:`STData` object with the updated
        defocus distances `defocus_fs` and `defocus_ss` for
        the fast and slow detector axes accordingly. Update
        `pixel_translations` based on the new defocus distances.

        Parameters
        ----------
        defocus_fs : float
            Defocus distance for the fast detector axis [m].
        defocus_ss : float, optional
            Defocus distance for the slow detector axis [m].
            Equals to `defocus_fs` if it's not provided.

        Returns
        -------
        STData
            New :class:`STData` object with the updated `defocus_ss`,
            `defocus_fs`, and `pixel_translations`.
        """
        if defocus_ss is None:
            defocus_ss = defocus_fs
        return {'defocus_ss': defocus_ss, 'defocus_fs': defocus_fs,
                'pixel_translations': None}

    def update_phase(self, st_obj):
        """Update `pixel_abberations`, `phase`, and `reference_image`
        based on :class:`SpeckleTracking` object `st_obj` data. `st_obj`
        must be derived from the :class:`STData` object, an error is
        raised otherwise.

        Parameters
        ----------
        st_obj : SpeckleTracking
            :class:`SpeckleTracking` object derived from the :class:`STData` object.

        Returns
        -------
        STData
            :class:`STData` object with the updated `pixel_abberations`,
            `phase`, and `reference_image`.

        Raises
        ------
        ValueError
            If `st_obj` doesn't belong to the :class:`STData` object.
        """
        if st_obj in self._st_objects:
            # Update phase, pixel_abberations, and reference_image
            dev_ss, dev_fs = (st_obj.pixel_map - self.get('pixel_map'))
            dev_ss -= dev_ss.mean()
            dev_fs -= dev_fs.mean()
            self.pixel_abberations = np.zeros(self.pixel_map.shape)
            self.pixel_abberations[:, self.roi[0]:self.roi[1], self.roi[2]:self.roi[3]] = np.stack((dev_ss, dev_fs))
            phase = ct_integrate(self.y_pixel_size**2 * dev_ss * self.defocus_ss / self.wavelength,
                                 self.x_pixel_size**2 * dev_fs * self.defocus_fs / self.wavelength) \
                    * 2 * np.pi / self.distance**2
            self.phase = np.zeros(self.whitefield.shape)
            self.phase[self.roi[0]:self.roi[1], self.roi[2]:self.roi[3]] = phase
            self.error_frame = np.zeros(self.whitefield.shape)
            self.error_frame[self.roi[0]:self.roi[1], self.roi[2]:self.roi[3]] = st_obj.error_frame
            self.reference_image = st_obj.reference_image

            # Initialize AbberationsFit objects
            self._ab_fits.clear()
            self._ab_fits.extend([AbberationsFit.import_data(self, axis=0),
                                  AbberationsFit.import_data(self, axis=1)])
            return self
        else:
            raise ValueError("the SpeckleTracking object doesn't belong to the data container")

    def fit_phase(self, axis=1, max_order=2, xtol=1e-14, ftol=1e-14, loss='cauchy'):
        """Fit `pixel_abberations` with the polynomial function
        using nonlinear least-squares algorithm. The function uses
        least-squares algorithm from :func:`scipy.optimize.least_squares`.

        Parameters
        ----------
        axis : int, optional
            Axis along which `pixel_abberations` is fitted.
        max_order : int, optional
            Maximum order of the polynomial model function.
        xtol : float, optional
            Tolerance for termination by the change of the independent variables.
        ftol : float, optional
            Tolerance for termination by the change of the cost function.
        loss : {'linear', 'soft_l1', 'huber', 'cauchy', 'arctan'}, optional
            Determines the loss function. The following keyword values are
            allowed:

            * 'linear' : ``rho(z) = z``. Gives a standard
              least-squares problem.
            * 'soft_l1' : ``rho(z) = 2 * ((1 + z)**0.5 - 1)``. The smooth
              approximation of l1 (absolute value) loss. Usually a good
              choice for robust least squares.
            * 'huber' : ``rho(z) = z if z <= 1 else 2*z**0.5 - 1``. Works
              similarly to 'soft_l1'.
            * 'cauchy' (default) : ``rho(z) = ln(1 + z)``. Severely weakens
              outliers influence, but may cause difficulties in optimization
              process.
            * 'arctan' : ``rho(z) = arctan(z)``. Limits a maximum loss on
              a single residual, has properties similar to 'cauchy'.

        Returns
        -------
        dict
            :class:`dict` with the following fields defined:

            * alpha : Third order abberations ceofficient [rad/mrad^3].
            * fit : Array of the polynomial function coefficients of the
              pixel abberations fit.
            * ph_fit : Array of the polynomial function coefficients of
              the phase abberations fit.
            * rel_err : Vector of relative errors of the fit coefficients.
            * r_sq : ``R**2`` goodness of fit.

        See Also
        --------
        AbberationsFit.fit - Full details of the abberations fitting algorithm.
        """
        if self._isphase:
            return self._ab_fits[axis].fit(max_order=max_order, xtol=xtol,
                                           ftol=ftol, loss=loss)
        else:
            return None

    def defocus_sweep(self, defoci_fs, defoci_ss=None, ls_ri=30, return_sweep=True):
        """Calculate a set of `reference_image` for each defocus in `defoci` and
        return a gradient magnitude for each `reference_image` as a figure of
        merit of it's sharpness (the higher the value the sharper `reference_image`
        is). `ls_ri` should be large enough in order to supress high frequency noise.
        Return a sweep image if `return_sweep` is True.

        Parameters
        ----------
        defoci_fs : numpy.ndarray
            Array of defocus distances along the fast detector axis [m].
        defoci_ss : numpy.ndarray, optional
            Array of defocus distances along the slow detector axis [m].
        ls_ri : float, optional
            `reference_image` length scale in pixels.
        return_sweep : bool, optional
            Return a sweep image if it's True.

        Returns
        -------
        grad_mag : numpy.ndarray
            Array of the average values of `reference_image` gradients squared.
        sweep_img : numpy.ndarray
            Defocus sweep image. Only if `return_sweep` is True.

        See Also
        --------
        SpeckleTracking.update_reference : `reference_image` update algorithm.
        """
        if defoci_ss is None:
            defoci_ss = defoci_fs.copy()
        grad_mag, sweep_scan = [], []
        for defocus_fs, defocus_ss in zip(defoci_fs.ravel(), defoci_ss.ravel()):
            st_data = self.update_defocus(defocus_fs, defocus_ss)
            st_obj = st_data.get_st().update_reference(ls_ri=ls_ri, sw_fs=0, sw_ss=0)
            ri_gm = gaussian_gradient_magnitude(st_obj.reference_image, sigma=ls_ri)
            sweep_scan.append(st_obj.reference_image)
            grad_mag.append(np.mean(ri_gm**2))
        grad_mag = np.array(grad_mag).reshape(defoci_fs.shape)
        if return_sweep:
            shape = tuple(np.max([ref_img.shape for ref_img in sweep_scan], axis=0))
            sweep_img = np.zeros((defoci_fs.shape + shape))
            for idx, ref_img in zip(np.ndindex(defoci_fs.shape), sweep_scan):
                sweep_img[idx][:ref_img.shape[0], :ref_img.shape[1]] = ref_img
            return grad_mag, sweep_img
        else:
            return grad_mag

    def get(self, attr, value=None):
        """Return a dataset with `mask` and `roi` applied.
        Return `value` if the attribute is not found.

        Parameters
        ----------
        attr : str
            Attribute to fetch.
        value : object, optional
            Return if `attr` is not found.

        Returns
        -------
        numpy.ndarray or object
            `attr` dataset with `mask` and `roi` applied.
            `value` if `attr` is not found.
        """
        if attr in self:
            val = super(STData, self).get(attr)
            if not val is None:
                if attr in ['data', 'error_frame', 'mask', 'phase',
                            'pixel_abberations', 'pixel_map', 'whitefield']:
                    val = val[..., self.roi[0]:self.roi[1], self.roi[2]:self.roi[3]]
                if attr in ['basis_vectors', 'data', 'mask', 'pixel_translations',
                            'translations']:
                    val = val[self.good_frames]
            return np.ascontiguousarray(val)
        else:
            return value

    def get_st(self):
        """Return :class:`SpeckleTracking` object derived
        from the container. Return None if `defocus_fs`
        or `defocus_ss` doesn't exist in the container.

        Returns
        -------
        SpeckleTracking or None
            An instance of :class:`SpeckleTracking` derived
            from the container. None if `defocus_fs` or
            `defocus_ss` is None.
        """
        if self._isdefocus:
            if self._st_objects:
                return self._st_objects[0]
            else:
                return SpeckleTracking.import_data(self)
        else:
            return None

    def get_st_list(self):
        """Return a list of all the :class:`SpeckleTracking`
        objects bound to the container.

        Returns
        -------
        list
            List of all the :class:`SpeckleTracking` objects
            bound to the container.
        """
        return self._st_objects

    def get_fit(self, axis=1):
        """Return an :class:`AbberationsFit` object for
        parametric regression of the lens' abberations
        profile. Return None if `defocus_fs` or
        `defocus_ss` doesn't exist in the container.

        Parameters
        ----------
        axis : int
            Detector axis along which the fitting is performed.

        Returns
        -------
        AbberationsFit or None
            An instance of :class:`AbberationsFit` class.
            None if `defocus_fs` or `defocus_ss` is None.
        """
        if self._isphase:
            return self._ab_fits[axis]
        else:
            return None

    def write_cxi(self, cxi_file, overwrite=True):
        """Write all the `attr` to a CXI file `cxi_file`.

        Parameters
        ----------
        cxi_file : h5py.File
            :class:`h5py.File` object of the CXI file.
        overwrite : bool, optional
            Overwrite the content of `cxi_file` file if it's True.

        Raises
        ------
        ValueError
            If `overwrite` is False and the data is already present
            in `cxi_file`.
        """
        for attr, data in self.items():
            if isinstance(data, np.ndarray):
                self.protocol.write_cxi(attr, data, cxi_file, overwrite=overwrite)

class SpeckleTracking(DataContainer):
    """Wrapper class for the  Robust Speckle Tracking algorithm.
    Performs `reference_image` and `pixel_map` updates.

    Parameters
    ----------
    **kwargs : dict
        Dictionary of the attributes' data specified in `attr_set`
        and `init_set`.

    Attributes
    ----------
    attr_set : set
        Set of attributes in the container which are necessary
        to initialize in the constructor.
    init_set : set
        Set of optional data attributes.

    Raises
    ------
    ValueError
        If an attribute specified in `attr_set` has not been provided.

    Notes
    -----
    Necessary attributes:

    * data : Measured frames.
    * dref : Reference to the :class:`STData` object, from which
      :class:`SpeckleTracking` object was derived.
    * dss_pix : The sample's translations along the slow axis
      in pixels.
    * dfs_pix : The sample's translations along the fast axis in pixels.
    * whitefield : Measured frames' whitefield.
    * pixel_map : The pixel mapping between the data at the detector's
      plane and the reference image at the reference plane.

    Optional attributes:

    * error_frame : MSE (mean-squared-error) of the reference image
      and pixel mapping fit per pixel.
    * reference_image : The unabberated reference image of the sample.
    * m0 : The lower bounds of the fast detector axis of
      the reference image at the reference frame in pixels.
    * n0 : The lower bounds of the slow detector axis of
      the reference image at the reference frame in pixels.

    See Also
    --------
    bin.make_reference : Full details of the `reference_image` update
        algorithm.
    bin.update_pixel_map_gs : Full details of the `pixel_map` update algorithm
        based on grid search.
    """
    attr_set = {'data', 'dref', 'dfs_pix', 'dss_pix', 'pixel_map', 'whitefield'}
    init_set = {'error_frame', 'ls_ri', 'n0', 'm0', 'reference_image'}

    def __init__(self, **kwargs):
        super(SpeckleTracking, self).__init__(**kwargs)
        self.dref._st_objects.append(self)

    @classmethod
    def import_data(cls, st_data):
        """Return a new :class:`SpeckleTracking` object
        with all the necessary data attributes imported from
        the :class:`STData` container object `st_data`.

        Parameters
        ----------
        st_data : STData
            :class:`STData` container object.

        Returns
        -------
        SpeckleTracking
            A new :class:`SpeckleTracking` object.
        """
        data = st_data.get('mask') * st_data.get('data')
        pixel_map = st_data.get('pixel_map')
        dij_pix = np.ascontiguousarray(np.swapaxes(st_data.get('pixel_translations'), 0, 1))
        return cls(data=data, dref=st_data, dfs_pix=dij_pix[1], dss_pix=dij_pix[0],
                   pixel_map=pixel_map, whitefield=st_data.get('whitefield'))

    @dict_to_object
    def update_reference(self, ls_ri,  sw_fs, sw_ss=0):
        """Return a new :class:`SpeckleTracking` object
        with the updated `reference_image`.

        Parameters
        ----------
        ls_ri : float, optional
            `reference_image` length scale in pixels.
        sw_ss : int, optional
            Search window size in pixels along the slow detector
            axis.
        sw_fs : int, optional
            Search window size in pixels along the fast detector
            axis.

        Returns
        -------
        SpeckleTracking
            A new :class:`SpeckleTracking` object with the updated
            `reference_image`.

        See Also
        --------
        bin.make_reference : Full details of the `reference_image` update
            algorithm.
        """
        I0, n0, m0 = make_reference(I_n=self.data, W=self.whitefield, u=self.pixel_map,
                                    di=self.dss_pix, dj=self.dfs_pix, sw_ss=sw_ss,
                                    sw_fs=sw_fs, ls=ls_ri)
        return {'ls_ri': ls_ri, 'n0': n0, 'm0': m0, 'reference_image': I0}

    @dict_to_object
    def update_pixel_map(self, ls_pm, sw_fs, sw_ss=0, method='search'):
        """Return a new :class:`SpeckleTracking` object with
        the updated `pixel_map`.

        Parameters
        ----------
        sw_fs : int
            Search window size in pixels along the fast detector
            axis.
        sw_ss : int, optional
            Search window size in pixels along the slow detector
            axis.
        ls_pm : float, optional
            `pixel_map` length scale in pixels.
        method : {'search', 'newton'}, optional
            `pixel_map` update algorithm. The following keyword
            values are allowed:

            * 'newton' : Iterative Newton's method based on
              finite difference gradient approximation.
            * 'search' : Grid search along the search window.

        Returns
        -------
        SpeckleTracking
            A new :class:`SpeckleTracking` object with the updated
            `pixel_map`.

        Raises
        ------
        AttributeError
            If `reference_image` was not generated beforehand.
        ValueError
            If `method` keyword value is not valid.

        See Also
        --------
        bin.update_pixel_map_gs, bin.update_pixel_map_nm : Full details
            of the `pixel_map` update algorithm.
        """
        if self.reference_image is None:
            raise AttributeError('The reference image has not been generated')
        else:
            if method == 'search':
                pixel_map = update_pixel_map_gs(I_n=self.data, W=self.whitefield,
                                                I0=self.reference_image, u0=self.pixel_map,
                                                di=self.dss_pix - self.n0,
                                                dj=self.dfs_pix - self.m0,
                                                sw_ss=sw_ss, sw_fs=sw_fs, ls=ls_pm)
            elif method == 'newton':
                pixel_map = update_pixel_map_nm(I_n=self.data, W=self.whitefield,
                                                I0=self.reference_image, u0=self.pixel_map,
                                                di=self.dss_pix - self.n0,
                                                dj=self.dfs_pix - self.m0,
                                                sw_fs=sw_fs, ls=ls_pm)
            else:
                raise ValueError('Wrong method argument: {:s}'.format(method))
            return {'pixel_map': pixel_map}
    
    @dict_to_object
    def update_errors(self):
        """Return a new :class:`SpeckleTracking` object with
        the updated `error_frame`.

        Parameters
        ----------
        ls_ri : float
            `reference_image` length scale in pixels.

        Returns
        -------
        SpeckleTracking
            A new :class:`SpeckleTracking` object with the updated
            `error_frame`.

        Raises
        ------
        AttributeError
            If `reference_image` was not generated beforehand.

        See Also
        --------
        bin.mse_frame : Full details of the error metric.
        """
        if self.reference_image is None:
            raise AttributeError('The reference image has not been generated')
        else:
            error_frame = mse_frame(I_n=self.data, W=self.whitefield, ls=self.ls_ri,
                                    I0=self.reference_image, u=self.pixel_map,
                                    di=self.dss_pix - self.n0, dj=self.dfs_pix - self.m0)
            return {'error_frame': error_frame}

    @dict_to_object
    def update_translations(self, sw_fs, sw_ss=0):
        """Return a new :class:`SpeckleTracking` object with
        the updated sample pixel translations (`dss_pix`,
        `dfs_pix`).

        Parameters
        ----------
        sw_fs : int
            Search window size in pixels along the fast detector
            axis.
        sw_ss : int, optional
            Search window size in pixels along the slow detector
            axis.
        ls_ri: float, optional
            `reference_image` length scale in pixels.
        method : {'search', 'newton'}, optional
            `pixel_map` update algorithm. The following keyword
            values are allowed:

            * 'newton' : Iterative Newton's method based on
              finite difference gradient approximation.
            * 'search' : Grid search along the search window.

        Returns
        -------
        SpeckleTracking
            A new :class:`SpeckleTracking` object with the updated
            `dss_pix`, `dfs_pix`.

        Raises
        ------
        AttributeError
            If `reference_image` was not generated beforehand.

        See Also
        --------
        bin.update_translations_gs : Full details of the sample
            translations update algorithm.
        """
        if self.reference_image is None:
            raise AttributeError('The reference image has not been generated')
        else:
            dij = update_translations_gs(I_n=self.data, W=self.whitefield,
                                         I0=self.reference_image, u=self.pixel_map,
                                         di=self.dss_pix - self.n0,
                                         dj=self.dfs_pix - self.m0,
                                         sw_ss=sw_ss, sw_fs=sw_fs, ls=self.ls_ri)
            dss_pix = np.ascontiguousarray(dij[:, 0]) + self.n0
            dfs_pix = np.ascontiguousarray(dij[:, 1]) + self.m0
            return {'dss_pix': dss_pix, 'dfs_pix': dfs_pix}

    def iter_update_gd(self, ls_ri, ls_pm, sw_fs, sw_ss=0, n_iter=30, f_tol=1e-6, momentum=0.,
                       learning_rate=1e1, gstep=.1, method='search', update_translations=False,
                       verbose=False, return_extra=False):
        """Perform iterative Robust Speckle Tracking update. `ls_ri` and
        `ls_pm` define high frequency cut-off to supress the noise. `ls_ri`
        is iteratively updated by dint of Gradient Descent with momentum
        update procedure. Iterative update terminates when the difference
        between total mean-squared-error (MSE) values of the two last iterations
        is less than `f_tol`.

        Parameters
        ----------
        sw_fs : int
            Search window size in pixels along the fast detector
            axis.
        sw_ss : int, optional
            Search window size in pixels along the slow detector
            axis.
        ls_pm : float, optional
            `pixel_map` length scale in pixels.
        ls_ri : float, optional
            Initial `reference_image` length scale in pixels.
        n_iter : int, optional
            Maximum number of iterations.
        f_tol : float, optional
            Tolerance for termination by the change of the total MSE.
        momentum : float, optional
            hyperparameter in interval (0.0-1.0) that accelerates gradient
            descent of `ls_ri` in the relevant direction and dampens oscillations.
        learning_rate : float, optional
            Learning rate for `ls_ri` update procedure.
        gstep : float, optional
            `ls_ri` step used to numerically calculate the MSE gradient.
        method : {'search', 'newton'}, optional
            `pixel_map` update algorithm. The following keyword
            values are allowed:

            * 'newton' : Iterative Newton's method based on
              finite difference gradient approximation.
            * 'search' : Grid search along the search window.
        update_translations : bool, optional
            Update sample pixel translations if True.
        return_extra : bool, optional
            Return errors and `ls_ri` array if True.

        Returns
        -------
        SpeckleTracking
            A new :class:`SpeckleTracking` object with the updated
            `pixel_map` and `reference_image`. `dss_pix` and `dfs_pix`
            are also updated if `update_translations` is True.
        list
            List of total MSE values for each iteration.  Only if
            `return_errors` is True.
        """
        velocity = 0.0
        obj = self.update_reference(ls_ri=ls_ri, sw_ss=sw_ss, sw_fs=sw_fs)
        obj.update_errors.inplace_update()
        extra = {'errors': [obj.error_frame.mean()],
                 'lss_ri': [ls_ri]}
        if verbose:
            print('Initial MSE = {:.6f}, Initial ls_ri = {:.2f}'.format(extra['errors'][-1],
                                                                        extra['lss_ri'][-1]))
        for it in range(1, n_iter + 1):
            # Update pixel_map
            new_obj = obj.update_pixel_map(ls_pm=ls_pm, sw_fs=sw_fs, sw_ss=sw_ss, method=method)
            obj.pixel_map += gaussian_filter(new_obj.pixel_map - obj.pixel_map,
                                             (0, ls_pm, ls_pm))

            # Update dss_pix, dfs_pix
            if update_translations:
                new_obj.update_translations.inplace_update(sw_ss=sw_ss, sw_fs=sw_fs)
                obj.dss_pix, obj.dfs_pix = new_obj.dss_pix, new_obj.dfs_pix

            # Update ls_ri
            grad = (obj.mse_total(ls_ri + gstep) - extra['errors'][-1]) / gstep
            velocity = np.clip(momentum * velocity - learning_rate * grad,
                               -0.75 * ls_ri, 0.75 * ls_ri)
            ls_ri += velocity
            extra['lss_ri'].append(ls_ri)

            # Update reference_image
            obj.update_reference.inplace_update(ls_ri=ls_ri, sw_fs=sw_fs, sw_ss=sw_ss)
            obj.update_errors.inplace_update()
            extra['errors'].append(obj.error_frame.mean())
            if verbose:
                print('Iteration No. {:d}: Total MSE = {:.6f}, ls_ri = {:.2f}'.format(it, extra['errors'][-1],
                                                                                      extra['lss_ri'][-1]))

            # Break if function tolerance is satisfied
            if (extra['errors'][-2] - extra['errors'][-1]) <= f_tol:
                break
        if return_extra:
            return obj, extra
        else:
            return obj

    def iter_update(self, ls_ri, ls_pm, sw_fs, sw_ss=0, n_iter=5, f_tol=1e-3,
                    method='search', update_translations=False, verbose=False, return_errors=False):
        """Perform iterative Robust Speckle Tracking update. `ls_ri` and
        `ls_pm` define high frequency cut-off to supress the noise.
        Iterative update terminates when the difference between total
        mean-squared-error (MSE) values of the two last iterations is
        less than `f_tol`.

        Parameters
        ----------
        sw_fs : int
            Search window size in pixels along the fast detector
            axis.
        sw_ss : int, optional
            Search window size in pixels along the slow detector
            axis.
        ls_pm : float, optional
            `pixel_map` length scale in pixels.
        ls_ri : float, optional
            `reference_image` length scale in pixels.
        n_iter : int, optional
            Maximum number of iterations.
        f_tol : float, optional
            Tolerance for termination by the change of the total MSE.
        method : {'search', 'newton'}, optional
            `pixel_map` update algorithm. The following keyword
            values are allowed:

            * 'newton' : Iterative Newton's method based on
              finite difference gradient approximation.
            * 'search' : Grid search along the search window.
        update_translations : bool, optional
            Update sample pixel translations if True.
        return_errors : bool, optional
            Return errors array if True.

        Returns
        -------
        SpeckleTracking
            A new :class:`SpeckleTracking` object with the updated
            `pixel_map` and `reference_image`. `dss_pix` and `dfs_pix`
            are also updated if `update_translations` is True.
        list
            List of total MSE values for each iteration.  Only if
            `return_errors` is True.
        """
        obj = self.update_reference(ls_ri=ls_ri, sw_ss=sw_ss, sw_fs=sw_fs)
        obj.update_errors.inplace_update()
        errors = [obj.error_frame.mean()]
        if verbose:
            print('Initial MSE = {:.6f}'.format(errors[0]))
        for it in range(1, n_iter + 1):
            # Update pixel_map
            new_obj = obj.update_pixel_map(ls_pm=ls_pm, sw_fs=sw_fs, sw_ss=sw_ss, method=method)
            obj.pixel_map += gaussian_filter(new_obj.pixel_map - obj.pixel_map,
                                             (0, ls_pm, ls_pm))

            # Update dss_pix, dfs_pix
            if update_translations:
                new_obj.update_translations.inplace_update(sw_ss=sw_ss, sw_fs=sw_fs)
                obj.dss_pix, obj.dfs_pix = new_obj.dss_pix, new_obj.dfs_pix

            # Update reference_image
            obj.update_reference.inplace_update(ls_ri=ls_ri, sw_fs=sw_fs, sw_ss=sw_ss)
            obj.update_errors.inplace_update()

            # Calculate errors
            errors.append(obj.error_frame.mean())
            if verbose:
                print('Iteration No. {:d}: Total MSE = {:.6f}'.format(it, errors[-1]))


            # Break if function tolerance is satisfied
            if (errors[-2] - errors[-1]) <= f_tol * errors[0]:
                break
        if return_errors:
            return obj, errors
        else:
            return obj

    def mse_total(self, ls_ri):
        """Generate a reference image with the given `ls_ri` and return
        average total mean-squared-error (MSE).

        Parameters
        ----------
        ls_ri : float
            `reference_image` length scale in pixels.

        Returns
        -------
        float
            Average total mean-squared-error (MSE).


        See Also
        --------
        bin.mse_total : Full details of the error metric.
        """
        I0, n0, m0 = make_reference(I_n=self.data, W=self.whitefield, u=self.pixel_map,
                                    di=self.dss_pix, dj=self.dfs_pix, sw_ss=0, sw_fs=0, ls=ls_ri)
        return mse_total(I_n=self.data, W=self.whitefield, I0=I0,
                         u=self.pixel_map, di=self.dss_pix - n0,
                         dj=self.dfs_pix - m0, ls=ls_ri)

    def mse_curve(self, lss_ri):
        """Return a mean-squared-error (MSE) survace.

        Parameters
        ----------
        lss_ri : numpy.ndarray
            Set of `reference_image` length scales in pixels.

        Returns
        -------
        numpy.ndarray
            A mean-squared-error (MSE) surface.

        See Also
        --------
        bin.mse_total : Full details of the error metric.
        """
        mse_list = []
        for ls_ri in np.array(lss_ri, ndmin=1):
            mse_list.append(self.mse_total(ls_ri))
        return np.array(mse_list)



"""st_sim provides wavefront propagation tools to simulate
Speckle Tracking scans. Wavefront propagation is based on
the Fresnel diffraction theory.
"""
from .st_sim_param import STParams, parameters
from .st_sim import STSim, STConverter, converter



cdef extern from "gsl/gsl_rng.h":

    ctypedef struct gsl_rng_type
    ctypedef struct gsl_rng

    cdef gsl_rng_type *gsl_rng_mt19937

    gsl_rng *gsl_rng_alloc ( gsl_rng_type * T) nogil

    unsigned long int gsl_rng_get ( gsl_rng * r) nogil

    void gsl_rng_set ( gsl_rng * r, unsigned long int seed) nogil

    void gsl_rng_free (gsl_rng * r) nogil

    double gsl_rng_uniform ( gsl_rng * r) nogil
    double gsl_rng_uniform_pos ( gsl_rng * r) nogil

cdef extern from "gsl/gsl_randist.h":

    unsigned int gsl_ran_poisson ( gsl_rng * r, double mu) nogil


cimport numpy as np
import numpy as np
from .beam_calc cimport *
from libc.math cimport sqrt, cos, sin, pi, erf
from libc.time cimport time_t, time
from libc.string cimport memcpy
from cython.parallel import prange
cimport openmp

cdef extern from "fft.c":

    int good_size_real(int n)
    int good_size_cmplx(int n)

    ctypedef struct cfft_plan_i
    cfft_plan_i* make_cfft_plan(int length) nogil
    void destroy_cfft_plan(cfft_plan_i* plan) nogil

    int cfft_forward(cfft_plan_i* plan, double* c, double fct) nogil
    int cfft_backward(cfft_plan_i* plan, double* c, double fct) nogil

    ctypedef struct rfft_plan_i
    rfft_plan_i* make_rfft_plan(int length) nogil
    void destroy_rfft_plan(rfft_plan_i* plan) nogil

    int rfft_forward(rfft_plan_i* plan, double* c, double fct) nogil
    int rfft_backward(rfft_plan_i* plan, double* c, double fct) nogil

ctypedef np.complex128_t complex_t
ctypedef np.uint64_t uint_t
ctypedef np.uint32_t uint32_t
ctypedef np.npy_bool bool_t

ctypedef fused numeric:
    np.float64_t
    np.float32_t
    np.uint64_t
    np.int64_t

DEF X_TOL = 4.320005384913445 # Y_TOL = 1e-9

cdef int fft(complex_t* arr, int n) nogil:
    cdef double* dptr = <double*>(arr)
    if not dptr:
        return -1
    cdef cfft_plan_i* plan = make_cfft_plan(n)
    if not plan:
        return -1
    cdef int fail = cfft_forward(plan, dptr, 1.0)
    if fail:
        return -1
    if plan:
        destroy_cfft_plan(plan)
    return 0

cdef int ifft(complex_t* arr, int n) nogil:
    cdef double* dptr = <double*>(arr)
    if not dptr:
        return -1
    cdef cfft_plan_i* plan = make_cfft_plan(n)
    if not plan:
        return -1
    cdef double fct = 1.0 / n
    cdef int fail = cfft_backward(plan, dptr, fct)
    if fail:
        return -1
    if plan:
        destroy_cfft_plan(plan)
    return 0

cdef int rfft(complex_t* res, double* arr, int n) nogil:
    cdef double* rptr = <double*>(res)
    cdef double* dptr = <double*>(arr)
    if not dptr:
        return -1
    cdef rfft_plan_i* plan = make_rfft_plan(n)
    if not plan:
        return -1
    memcpy(<char *>(rptr + 1), dptr, n * sizeof(double))
    cdef int fail = rfft_forward(plan, rptr + 1, 1.0)
    rptr[0] = rptr[1]; rptr[1] = 0.0
    if not n % 2:
        rptr[n + 1] = 0.0
    if fail:
        return -1
    if plan:
        destroy_rfft_plan(plan)
    return 0

cdef int irfft(double* res, complex_t* arr, int n) nogil:
    cdef double* dptr = <double*>(arr)
    cdef double* rptr = <double*>(res)
    if not dptr:
        return -1
    cdef rfft_plan_i* plan = make_rfft_plan(n)
    if not plan:
        return -1
    cdef double fct = 1.0 / n
    memcpy(<char *>(rptr + 1), dptr + 2, (n - 1) * sizeof(double))
    rptr[0] = dptr[0]
    cdef int fail = rfft_backward(plan, rptr, fct)
    if fail:
        return -1
    if plan:
        destroy_rfft_plan(plan)
    return 0

cdef void rsc_type1_c(complex_t[::1] u, complex_t[::1] k, complex_t[::1] h,
                      double dx0, double dx, double z, double wl) nogil:
    cdef:
        int n = u.shape[0], i
        double ph, dist
        complex_t u0
    for i in range(n):
        dist = (dx0 * (i - n // 2))**2 + z**2
        ph = 2 * pi / wl * sqrt(dist)
        k[i] = -dx0 * z / sqrt(wl) * (sin(ph) + 1j * cos(ph)) / dist**0.75
    fft(&u[0], n)
    fft(&k[0], n)
    for i in range(n):
        ph = pi * (<double>(i) / n - (2 * i) // n)**2 * dx / dx0 * n
        u[i] = u[i] * k[i] * (cos(ph) - 1j * sin(ph))
        k[i] = cos(ph) - 1j * sin(ph)
        h[i] = cos(ph) + 1j * sin(ph)
    ifft(&u[0], n)
    ifft(&h[0], n)
    for i in range(n):
        u[i] = u[i] * h[i]
    fft(&u[0], n)
    for i in range(n // 2):
        u0 = u[i] * k[i]; u[i] = u[i + n // 2] * k[i + n // 2]
        u[i + n // 2] = u0

cdef void rsc_type2_c(complex_t[::1] u, complex_t[::1] k, complex_t[::1] h,
                      double dx0, double dx, double z, double wl) nogil:
    cdef:
        int n = u.shape[0], i
        double ph, ph2, dist
    for i in range(n):
        dist = (dx * (i - n // 2))**2 + z**2
        ph = 2 * pi / wl * sqrt(dist)
        k[i] = -dx0 * z / sqrt(wl) * (sin(ph) + 1j * cos(ph)) / dist**0.75
    fft(&k[0], n)
    for i in range(n):
        ph = pi * (<double>(i) / n - (2 * i) // n)**2 * dx0 / dx * n
        ph2 = pi * (i - n // 2)**2 * dx0 / dx / n
        u[i] = u[i] * (cos(ph2) + 1j * sin(ph2))
        k[i] = k[i] * (cos(ph) + 1j * sin(ph))
        h[i] = cos(ph2) - 1j * sin(ph2)
    fft(&u[0], n)
    fft(&h[0], n)
    for i in range(n):
        u[i] = u[i] * h[i]
    ifft(&u[0], n)
    for i in range(n):
        u[i] = u[i] * k[i]
    ifft(&u[0], n)

cdef void rsc_wp_c(complex_t[::1] u, complex_t[::1] k, complex_t[::1] h, complex_t[::1] u0,
                   double dx0, double dx, double z, double wl) nogil:
    cdef:
        int a = u0.shape[0], n = u.shape[0], i
    for i in range((n - a) // 2 + 1):
        u[i] = 0; u[n - 1 - i] = 0
    for i in range(a):
        u[i + (n - a) // 2] = u0[i]
    if dx0 >= dx:
        rsc_type1_c(u, k, h, dx0, dx, z, wl)
    else:
        rsc_type2_c(u, k, h, dx0, dx, z, wl)

def rsc_wp(complex_t[::1] u0, double dx0, double dx, double z, double wl):
    r"""Wavefront propagator based on Rayleigh-Sommerfeld convolution
    method [RSC]_. Propagates a wavefront `u0` by `z` distance
    downstream.

    Parameters
    ----------
    u0 : numpy.ndarray
        Initial wavefront.
    dx0 : float
        Sampling interval at the plane upstream [um].
    dx : float
        Sampling interval at the plane downstream [um].
    z : float
        Propagation distance [um].
    wl : float
        Incoming beam's wavelength [um].

    Returns
    -------
    u : numpy.ndarray
        Propagated wavefront.

    Notes
    -----
    The Rayleigh–Sommerfeld diffraction integral transform is defined as:

    .. math::
        u^{\prime}(x^{\prime}) = \frac{z}{j \sqrt{\lambda}} \int_{-\infty}^{+\infty}
        u(x) \mathrm{exp} \left[-j k r(x, x^{\prime}) \right] dx
    
    with

    .. math::
        r(x, x^{\prime}) = \left[ (x - x^{\prime})^2 + z^2 \right]^{1 / 2}

    References
    ----------
    .. [RSC] V. Nascov and P. C. Logofătu, "Fast computation algorithm
             for the Rayleigh-Sommerfeld diffraction formula using
             a type of scaled convolution," Appl. Opt. 48, 4310-4319
             (2009).
    """
    cdef:
        int a = u0.shape[0], i
        double alpha = dx0 / dx if dx0 <= dx else dx / dx0
        int n = good_size_cmplx(2 * (<int>(a * (1 + alpha) // 2) + 1))
        complex_t[::1] u = np.empty(n, dtype=np.complex128)
        complex_t[::1] k = np.empty(n, dtype=np.complex128)
        complex_t[::1] h = np.empty(n, dtype=np.complex128)
    rsc_wp_c(u, k, h, u0, dx0, dx, z, wl)
    return np.asarray(u[(n - a) // 2:(n + a) // 2])

cdef void fhf_wp_c(complex_t[::1] u, complex_t[::1] k, complex_t[::1] u0,
                   double dx0, double dx, double z, double wl) nogil:
    cdef:
        int a = u0.shape[0], n = u.shape[0], i
        double ph0 = 2 * pi / wl * z, ph1
        double alpha = dx0 * dx / wl / z
        complex_t h0 = -(sin(ph0) + 1j * cos(ph0)) / sqrt(wl * z) * dx0, w0, w1
    for i in range((n - a) // 2 + 1):
        u[i] = 0; u[n - 1 - i] = 0
    for i in range(a):
        u[i + (n - a) // 2] = u0[i]
    for i in range(n):
        ph0 = pi * (i - n // 2)**2 * alpha
        k[i] = cos(ph0) - 1j * sin(ph0)
        u[i] = u[i] * (cos(ph0) + 1j * sin(ph0))
    fft(&u[0], n)
    fft(&k[0], n)
    for i in range(n):
        u[i] = u[i] * k[i]
    ifft(&u[0], n)
    for i in range(n // 2):
        ph0 = pi * i**2 * alpha
        ph1 = pi * (i - n // 2)**2 * alpha
        w0 = (cos(ph0) - 1j * sin(ph0)) * u[i]
        w1 = (cos(ph1) - 1j * sin(ph1)) * u[i + n // 2]
        u[i] = h0 * (cos(ph1 / dx0 * dx) - 1j * sin(ph1 / dx0 * dx)) * w1
        u[i + n // 2] = h0 * (cos(ph0 / dx0 * dx) - 1j * sin(ph0 / dx0 * dx)) * w0

def fhf_wp(complex_t[::1] u0, double dx0, double dx, double z, double wl):
    r"""One-dimensional discrete form of Fraunhofer diffraction
    performed by the means of Fast Fourier transform.

    Parameters
    ----------
    u0 : numpy.ndarray
        Wavefront at the plane upstream.
    dx0 : float
        Sampling interval at the plane upstream [um].
    dx : float
        Sampling interval at the plane downstream [um].
    z : float
        Propagation distance [um].
    wl : float
        Incoming beam's wavelength [um].

    Returns
    -------
    numpy.ndarray
        Wavefront at the plane downstream.

    Notes
    -----
    The Fraunhofer integral transform is defined as:

    .. math::
        u^{\prime}(x^{\prime}) = \frac{e^{-j k z}}{j \sqrt{\lambda z}}
        e^{-\frac{j k}{2 z} x^{\prime 2}} \int_{-\infty}^{+\infty} u(x)
        e^{j\frac{2 \pi}{\lambda z} x x^{\prime}} dx
    """
    cdef:
        int a = u0.shape[0]
        int n = good_size_cmplx(2 * a - 1)
        complex_t[::1] u = np.empty(n, dtype=np.complex128)
        complex_t[::1] k = np.empty(n, dtype=np.complex128)
    fhf_wp_c(u, k, u0, dx0, dx, z, wl)
    return np.asarray(u[(n - a) // 2:(n + a) // 2])

def fhf_wp_scan(complex_t[:, ::1] u0, double dx0, double dx, double z, double wl):
    """One-dimensional discrete form of Fraunhofer diffraction
    performed by the means of Fast Fourier transform. The transform
    is applied to a set of wavefronts `u0`.

    Parameters
    ----------
    u0 : numpy.ndarray
        Set of wavefronts at the plane upstream.
    dx0 : float
        Sampling interval at the plane upstream [um].
    dx : float
        Sampling interval at the plane downstream [um].
    z : float
        Propagation distance [um].
    wl : float
        Incoming beam's wavelength [um].

    Returns
    -------
    numpy.ndarray
        Set of wavefronts at the plane downstream.

    See Also
    --------
    fhf_wp : Description of the Fraunhofer diffraction
        integral transform.
    """
    cdef:
        int a = u0.shape[0], b = u0.shape[1], i, t
        int n = good_size_cmplx(2 * b - 1)
        int max_threads = openmp.omp_get_max_threads()
        complex_t[:, ::1] u = np.empty((a, n), dtype=np.complex128)
        complex_t[:, ::1] k = np.empty((max_threads, n), dtype=np.complex128)
    for i in prange(a, schedule='guided', nogil=True):
        t = openmp.omp_get_thread_num()
        fhf_wp_c(u[i], k[t], u0[i], dx0, dx, z, wl)
    return np.asarray(u[:, (n - b) // 2:(n + b) // 2], order='C')

def bar_positions(double x0, double x1, double b_dx, double rd):
    """Generate a coordinate array of randomized barcode's bar positions.

    Parameters
    ----------
    x0 : float
        Barcode's lower bound along the x axis [um].
    x1 : float
        Barcode's upper bound along the x axis [um].
    b_dx : float
        Average bar's size [um].
    rd : float
        Random deviation of barcode's bar positions (0.0 - 1.0).

    Returns
    -------
    bx_arr : numpy.ndarray
        Array of barcode's bar coordinates.
    """
    cdef:
        int br_n = 2 * (<int>((x1 - x0) / 2 / b_dx) + 1) if x1 > x0 else 0, i
        gsl_rng *r = gsl_rng_alloc(gsl_rng_mt19937)
        double[::1] bar_pos = np.empty(br_n, dtype=np.float64)
        time_t t = time(NULL)
    gsl_rng_set(r, t)
    if br_n:
        for i in range(br_n):
            bar_pos[i] = x0 + b_dx * (i + 2 * rd * (gsl_rng_uniform_pos(r) - 0.5))
    gsl_rng_free(r)
    return np.asarray(bar_pos)

cdef int binary_search(double[::1] values, int l, int r, double x) nogil:
    cdef int m = l + (r - l) // 2
    if l <= r:
        if x == values[m]:
            return m
        elif x > values[m] and x <= values[m + 1]:
            return m + 1
        elif x < values[m]:
            return binary_search(values, l, m, x)
        else:
            return binary_search(values, m + 1, r, x)

cdef int searchsorted(double[::1] values, double x) nogil:
    cdef int r = values.shape[0]
    if x < values[0]:
        return 0
    elif x > values[r - 1]:
        return r
    else:
        return binary_search(values, 0, r, x)

cdef double barcode_c(double[::1] bar_pos, double xx, double b_atn, double b_sgm, double blk_atn) nogil:
    cdef:
        int b = bar_pos.shape[0], i, j0, j
        double b_dx = (bar_pos[b - 1] - bar_pos[0]) / b
        int bb = <int>(X_TOL * sqrt(2) * b_sgm / b_dx + 1)
        double tr, x0, x1
    j0 = searchsorted(bar_pos, xx) # even '-', odd '+'
    tr = 0
    for j in range(j0 - bb, j0 + bb + 1):
        if j > 0 and j < b - 1:
            x0 = (xx - bar_pos[j - 1]) / sqrt(2) / b_sgm
            x1 = (xx - bar_pos[j]) / sqrt(2) / b_sgm
            tr += b_atn * (bar_pos[j] - bar_pos[j - 1]) / b_dx * 0.5 * (0.5 - j % 2) * (erf(x0) - erf(x1))
    tr -= (0.25 * b_atn + 0.5 * blk_atn) * erf((xx - bar_pos[0]) / sqrt(2 + 2 * (blk_atn / b_atn)**2) / b_sgm)
    tr += (0.25 * b_atn + 0.5 * blk_atn) * erf((xx - bar_pos[b - 1]) / sqrt(2 + 2 * (blk_atn / b_atn)**2) / b_sgm)
    return sqrt(1 + tr)
        
def barcode_profile(double[::1] bar_pos, double[::1] x_arr, double b_atn, double b_sgm, double blk_atn):
    r"""Return an array of barcode's transmission profile calculated
    at `x_arr` coordinates.

    Parameters
    ----------
    bar_pos : numpy.ndarray
        Coordinates of barcode's bar positions [um].
    x_arr : numpy.ndarray
        Array of the coordinates, where the transmission coefficients
        are calculated [um].    
    b_atn : float
        Barcode's bar attenuation coefficient (0.0 - 1.0).
    b_sgm : float
        Bar's blurriness [um].
    blk_atn : float
        Barcode's bulk attenuation coefficient (0.0 - 1.0).
    
    Returns
    -------
    b_tr : numpy.ndarray
        Array of barcode's transmission profiles.

    Notes
    -----
    Barcode's transmission profile is simulated with a set
    of error functions:
    
    .. math::
        \begin{multline}
            T_{b}(x) = 1 - \frac{T_{bulk}}{2} \left\{
            \mathrm{erf}\left[ \frac{x - x_{bar}[0]}{\sqrt{2} \sigma} \right] +
            \mathrm{erf}\left[ \frac{x_{bar}[n - 1] - x}{\sqrt{2} \sigma} \right]
            \right\} -\\
            \frac{T_{bar}}{4} \sum_{i = 1}^{n - 2} \left\{
            2 \mathrm{erf}\left[ \frac{x - x_{bar}[i]}{\sqrt{2} \sigma} \right] -
            \mathrm{erf}\left[ \frac{x - x_{bar}[i - 1]}{\sqrt{2} \sigma} \right] -
            \mathrm{erf}\left[ \frac{x - x_{bar}[i + 1]}{\sqrt{2} \sigma} \right]
            \right\}
        \end{multline}
    
    where :math:`x_{bar}` is an array of bar coordinates.
    """
    cdef:
        int a = x_arr.shape[0], i
        double[::1] b_tr = np.empty(a, dtype=np.float64)
    for i in prange(a, schedule='guided', nogil=True):
        b_tr[i] = barcode_c(bar_pos, x_arr[i], b_atn, b_sgm, blk_atn)
    return np.asarray(b_tr)

cdef void fft_convolve_c(double[::1] b1, complex_t[::1] c1, double[::1] a1, complex_t[::1] c2) nogil:
    cdef:
        int a = a1.shape[0], n = b1.shape[0], i
        double b0
    for i in range((n - a) // 2 + 1):
        b1[i] = a1[0]; b1[n - 1 - i] = a1[a - 1]
    for i in range(a):
        b1[i + (n - a) // 2] = a1[i]
    rfft(&c1[0], &b1[0], n)
    for i in range(n // 2 + 1):
        c1[i] = c1[i] * c2[i]
    irfft(&b1[0], &c1[0], n)
    for i in range(n // 2):
        b0 = b1[i]; b1[i] = b1[i + n // 2]
        b1[i + n // 2] = b0

def fft_convolve(double[::1] a1, double[::1] a2):
    """Convolve two one-dimensional arrays using FFT. The output
    size is the same size as `a1`.

    Parameters
    ----------
    a1 : numpy.ndarray
        First input array.
    a2 : numpy.ndarray
        Second input array.

    Returns
    -------
    out : numpy.ndarray
        A one-dimensional array containing the discrete linear
        convolution of `a1` with `a2`.
    """
    cdef:
        int a = a1.shape[0], b = a2.shape[0], i
        int n = good_size_real(a + b - 1)
        double[::1] b1 = np.empty(n, dtype=np.float64)
        double[::1] b2 = np.empty(n, dtype=np.float64)
        complex_t[::1] c1 = np.empty(n // 2 + 1, dtype=np.complex128)
        complex_t[::1] c2 = np.empty(n // 2 + 1, dtype=np.complex128)
    for i in range((n - b) // 2 + 1):
        b2[i] = a2[0]; b2[n - 1 - i] = a2[b - 1]
    for i in range(b):
        b2[i + (n - b) // 2] = a2[i]
    rfft(&c2[0], &b2[0], n)
    fft_convolve_c(b1, c1, a1, c2)
    return np.asarray(b1[(n - a) // 2:(n + a) // 2])

def fft_convolve_scan(double[:, ::1] a1, double[::1] a2):
    """Convolve a set of one-dimensional arrays `a1` with `a2` using
    FFT. The output size is the same size as `a1`.

    Parameters
    ----------
    a1 : numpy.ndarray
        Set of input arrays.
    a2 : numpy.ndarray
        Second input array.

    Returns
    -------
    out : numpy.ndarray
        A set of one-dimensional arrays containing the discrete
        linear convolution of `a1` with `a2`.
    """
    cdef:
        int nf = a1.shape[0], a = a1.shape[1], b = a2.shape[0], i, t
        int n = good_size_real(a + b - 1)
        int max_threads = openmp.omp_get_max_threads()
        double[:, ::1] b1 = np.empty((nf, n), dtype=np.float64)
        double[::1] b2 = np.empty(n, dtype=np.float64)
        complex_t[:, ::1] c1 = np.empty((max_threads, n // 2 + 1), dtype=np.complex128)
        complex_t[::1] c2 = np.empty(n // 2 + 1, dtype=np.complex128)
    for i in range((n - b) // 2 + 1):
        b2[i] = a2[0]; b2[n - 1 - i] = a2[b - 1]
    for i in range(b):
        b2[i + (n - b) // 2] = a2[i]
    rfft(&c2[0], &b2[0], n)
    for i in prange(nf, schedule='guided', nogil=True):
        t = openmp.omp_get_thread_num()
        fft_convolve_c(b1[i], c1[t], a1[i], c2)
    return np.asarray(b1[:, (n - a) // 2:(n + a) // 2], order='C')

cdef void make_frame_c(double[:, ::1] frame, double[::1] i_x, double[::1] i_ss, double dx) nogil:
    cdef:
        int a = i_x.shape[0], ss = frame.shape[0], fs = frame.shape[1], i, j, jj, j0, j1
        double i_fs = 0
    j1 = <int>(0.5 * a / fs)
    for jj in range(j1):
        i_fs += i_x[jj] * dx
    for i in range(ss):
        frame[i, 0] = <int>(i_fs * i_ss[i])
    i_fs = 0
    for jj in range(j1):
        i_fs += i_x[a - 1 - jj] * dx
    for i in range(ss):
        frame[i, fs - 1] = <int>(i_fs * i_ss[i])
    for j in range(1, fs - 1):
        i_fs = 0
        j0 = <int>((j - 0.5) * a // fs)
        j1 = <int>((j + 0.5) * a // fs)
        for jj in range(j0, j1):
            i_fs += i_x[jj] * dx
        for i in range(ss):
            frame[i, j] = i_fs * i_ss[i]

def make_frames(double[:, ::1] i_x, double[::1] i_y, double dx, double dy, int ss, int fs):
    """Generate intensity frames with Poisson noise
    from one-dimensional intensity profiles `i_x`
    and `i_y` convoluted with the source's rocking
    curves `sc_x` and `sc_y`.

    Parameters
    ----------
    i_x : numpy.ndarray
        Intensity profile along the x axis.
    i_y : numpy.ndarray
        Intensity profile along the y axis.
    dx : float
        Sampling interval along the x axis [um].
    dy : float
        Sampling interval along the y axis [um].
    ss : int
        Detector's size along the slow axis.
    fs : int
        Detector's size along the fast axis.
    apply_noise : bool
        Apply Poisson noise if it's True.

    Returns
    -------
    frames : numpy.ndarray
        Intensity frames.
    """
    cdef:
        int nf = i_x.shape[0], a = i_x.shape[1], b = i_y.shape[0], i, ii, i0, i1
        int max_threads = openmp.omp_get_max_threads(), t
        double[:, :, ::1] frames = np.zeros((nf, ss, fs), dtype=np.float64)
        double[::1] i_ss = np.zeros(ss, dtype=np.float64)
    i1 = <int>(0.5 * b // ss)
    for ii in range(i1):
        i_ss[0] += i_y[ii] * dy
        i_ss[ss - 1] += i_y[b - 1 - ii] * dy
    for i in range(1, ss - 1):
        i0 = <int>((i - 0.5) * b // ss)
        i1 = <int>((i + 0.5) * b // ss)
        for ii in range(i0, i1):
            i_ss[i] += i_y[ii] * dy
    for i in prange(nf, schedule='guided', nogil=True):
        make_frame_c(frames[i], i_x[i], i_ss, dx)
    return np.asarray(frames)

cdef uint32_t noisy_frame_c(uint32_t[:, ::1] noisy, double[:, ::1] frame, uint32_t seed) nogil:
    cdef:
        int ss = frame.shape[0], fs = frame.shape[1], i, j
        gsl_rng *r = gsl_rng_alloc(gsl_rng_mt19937)
    gsl_rng_set(r, seed)
    for i in range(ss):
        for j in range(fs):
            noisy[i, j] = gsl_ran_poisson(r, frame[i, j])
    seed = gsl_rng_get(r)
    gsl_rng_free(r)
    return seed

def apply_poisson(double[:, :, ::1] frames):
    cdef:
        int nf = frames.shape[0], ss = frames.shape[1], fs = frames.shape[2], i
        int max_threads = openmp.omp_get_max_threads(), t
        gsl_rng *r = gsl_rng_alloc(gsl_rng_mt19937)
        time_t tm = time(NULL)
        uint32_t[::1] seeds = np.empty(max_threads, dtype=np.uint32)
        uint32_t[:, :, ::1] noisy = np.empty((nf, ss, fs), dtype=np.uint32)
    gsl_rng_set(r, tm)
    for i in range(max_threads):
        seeds[i] = gsl_rng_get(r)
    for i in prange(nf, schedule='guided', nogil=True):
        t = openmp.omp_get_thread_num()
        seeds[t] = noisy_frame_c(noisy[i], frames[i], seeds[t])
    gsl_rng_free(r)
    return np.asarray(noisy)

cdef numeric wirthselect(numeric[::1] array, int k) nogil:
    cdef:
        int l = 0, m = array.shape[0] - 1, i, j
        numeric x, tmp 
    while l < m: 
        x = array[k] 
        i = l; j = m 
        while 1: 
            while array[i] < x: i += 1 
            while x < array[j]: j -= 1 
            if i <= j: 
                tmp = array[i]; array[i] = array[j]; array[j] = tmp
                i += 1; j -= 1 
            if i > j: break 
        if j < k: l = i 
        if k < i: m = j 
    return array[k]

def make_whitefield(numeric[:, :, ::1] data, bool_t[:, :, ::1] mask):
    """Generate a whitefield using the median filtering.

    Parameters
    ----------
    data : numpy.ndarray
        Intensity frames.
    mask : numpy.ndarray
        Bad pixel mask.

    Returns
    -------
    wf : numpy.ndarray
        Whitefield.
    """
    if numeric is np.float64_t:
        dtype = np.float64
    elif numeric is np.float32_t:
        dtype = np.float32
    else:
        dtype = np.uint64
    cdef:
        int a = data.shape[0], b = data.shape[1], c = data.shape[2], t, i, j, k, ii
        int max_threads = openmp.omp_get_max_threads()
        numeric[:, ::1] wf = np.empty((b, c), dtype=dtype)
        numeric[:, ::1] array = np.empty((max_threads, a), dtype=dtype)
    for j in prange(b, schedule='guided', nogil=True):
        t = openmp.omp_get_thread_num()
        for k in range(c):
            ii = 0
            for i in range(a):
                if mask[i, j, k]:
                    array[t, ii] = data[i, j, k]
                    ii = ii + 1
            wf[j, k] = wirthselect(array[t], ii // 2)
    return np.asarray(wf)



"""rst is a Python library for wavefront metrology and sample imaging
based on ptychographic speckle tracking algorithm. This project
takes over Andrew Morgan's `Ptychographic Speckle Tracking project`_
as an improved version aiming to add robustness to the optimisation
algorithm in the case of the high noise present in the measured data.

.. _Ptychographic Speckle Tracking project: https://github.com/andyofmelbourne/speckle-tracking

(c) Nikolay Ivanov, 2020.
"""
from __future__ import absolute_import

from .data_processing import SpeckleTracking, STData
from .abberations_fit import AbberationsFit
from .protocol import INIParser, Protocol, STLoader, loader, cxi_protocol
from .bin import *
